""".. module:: emulsion.tools.functions

Additional functions for symbolic computing in YAML model definition files.

All functions in this module can be used in EMULSION models.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import numpy as np

def IfThenElse(condition, val_if_true, val_if_false):
    """Ternary conditional function: return either *val_if_true* or
    *val_if_false* depending on *condition*.

    Example
    -------
    Here in a parameter definition (assuming that ``summer_period`` is
    defined e.g. in the *calendars* section of the model)::

      average_temperature:
        desc: 'average temperature for the season'
        value: 'IfThenElse(summer_period, 25, 8)'

    Parameters
    ----------
    condition: bool
        a boolean expression
    val_if_true: number
        value to return if the expression is True
    val_if_false: number
        value to return if the expression is False

    Returns
    -------
    number:
        One of the two values, depending on *condition*.

    """
    return val_if_true if condition else val_if_false

IfThenElse.__USER_FUNCTION__ = ['Functions Available for Models']


# BEWARE: using lambdify, by default And and Or fall back to numpy's
# *binary* operators, so that more than 3 conditions linked by And or
# Or trigger a TypeError ('return arrays must be of ArrayType')
# see topic addressed there:
# https://stackoverflow.com/questions/42045906/typeerror-return-arrays-must-be-of-arraytype-using-lambdify-of-sympy-in-python

# To avoid problems, please use AND / OR (fully in UPPERCASE) instead
# of And / Or (Capitalized) in the conditions of Emulsion models
def AND(*values):
    """Return a logical AND (conjunction) between all the values.

    Example
    -------
    To define symptomatic individuals as infected for at least 5 days::

      is_symptomatic:
        desc: 'test if individuals has symptoms'
        value: 'AND(is_I, duration_in_health_state >= 5)'

    Parameters
    ----------
    *values: list
        boolean values separated by commas

    Returns
    -------
    bool:
        True if **all** values are True, False otherwise.
    """
    return all(values)

AND.__USER_FUNCTION__ = ['Functions Available for Models']


def OR(*values):
    """Return a logical OR (disjunction) between all the values.

    Example
    -------
    To sell animals depending on either their weight or their age::

      transitions:
        ...
        - from: Fattening
          to: Sold
          proba: 1
          cond: 'OR(weight >= weight_thr, age >= age_thr)'

    Parameters
    ----------
    *values: list
        boolean values separated by commas

    Returns
    -------
    bool:
        True if **at least one** of the values is True, False otherwise.
    """
    return any(values)

OR.__USER_FUNCTION__ = ['Functions Available for Models']

def MIN(*values):
    """Return a the minimum of all values.

    Parameters
    ----------
    *values: list
        boolean values separated by commas

    Returns
    -------
    float:
        the lowest value
    """
    return min(values)

MIN.__USER_FUNCTION__ = ['Functions Available for Models']

def MAX(*values):
    """Return a the maximum of all values.

    Parameters
    ----------
    *values: list
        boolean values separated by commas

    Returns
    -------
    float:
        the highest value
    """
    return max(values)

MAX.__USER_FUNCTION__ = ['Functions Available for Models']



def random_bool(proba_success: float) -> int:
    """Return a random boolean value (actually, 0 or 1) depending on
    *proba_success*.

    Example
    -------
    Set the value of a state variable ``has_symptoms`` when entering
    in the infectious state::

      states:
        ...
        - I:
          ...
          on_enter:
            - set_var: has_symptoms
              value: 'random_bool(proba_symptomatic)'

    Parameters
    ----------
    proba_success: float in [0,1]
        probability of returning 1 (True)

    Returns
    -------
    int:
        either 1 with probability *proba_success*, or 0 with
        probability 1-*proba_success*

    """
    return np.random.binomial(1, proba_success)

random_bool.__USER_FUNCTION__ = ['Functions Available for Models']


def random_choice(*values):
    """Return a value chosen randomly among those provided (equiprobable
    sampling).

    Example
    -------
    To init ``age`` among three typical values::

      prototypes:
        individuals:
          init_individual:
            age: 'random_choice(10, 50, 200)'

    Parameters
    ----------
    *values: list
        the possible values, separated by commas

    Returns
    -------
    val:
        one of the values (equiprobable choice)

    """
    return np.random.choice(values)

random_choice.__USER_FUNCTION__ = ['Functions Available for Models']


def random_choice_weighted(*values):
    """Return a value chosen randomly among those provided (but not
    equiprobably).

    Example
    -------
    To init ``age`` among three typical values which respectively
    represent 10%, 20% and 70% of the population::

      prototypes:
        individuals:
          init_individual:
            age: 'random_choice_weighted(10, 0.1, 50, 0.2, 200, 0.7)'

    Parameters
    ----------
    values: list
        possibles choices and their weight, alternatively, separated
        by commas. Weights are normalized to be used as probabilities.

    Returns
    -------
    One of the choices

    """
    choices_and_values = zip(*[iter(values)]*2)
    choices, weights = zip(*choices_and_values)
    total = sum(weights)
    probas = [w / total for w in weights]
    return np.random.choice(choices, p=probas)

random_choice_weighted.__USER_FUNCTION__ = ['Functions Available for Models']


def random_multinomial(number_of_samples, *probas):
    """Return a multinomial sample based on the specified probabilities.

    Parameters
    ----------
    number_of_samples: int
        number of experiments
    probas: list
        list of probabilities

    Returns
    -------
    list:
        the drawn samples
    """
    return list(np.random.multinomial(number_of_samples, probas))

random_multinomial.__USER_FUNCTION__ = ['Functions Available for Models']


#   _____ _                _             _         _
#  / ____| |              | |           | |       | |
# | (___ | |__   ___  _ __| |_ ___ _   _| |_ ___  | |_ ___
#  \___ \| '_ \ / _ \| '__| __/ __| | | | __/ __| | __/ _ \
#  ____) | | | | (_) | |  | || (__| |_| | |_\__ \ | || (_) |
# |_____/|_| |_|\___/|_|   \__\___|\__,_|\__|___/  \__\___/
#                                                          _
#                                                         | |
#  _ __  _   _ _ __ ___  _ __  _   _   _ __ __ _ _ __   __| | ___  _ __ ___
# | '_ \| | | | '_ ` _ \| '_ \| | | | | '__/ _` | '_ \ / _` |/ _ \| '_ ` _ \
# | | | | |_| | | | | | | |_) | |_| |_| | | (_| | | | | (_| | (_) | | | | | |
# |_| |_|\__,_|_| |_| |_| .__/ \__, (_)_|  \__,_|_| |_|\__,_|\___/|_| |_| |_|
#                       | |     __/ |
#                       |_|    |___/

def random_uniform(low: float, high: float) -> float:
    """Return a random value drawn from a uniform distribution between
    *low* (inclusive) and *high* (exclusive).

    Example
    -------
    In a prototype definition::

      age: random_uniform(min_age, max_age)

    Parameters
    ----------
    low: float
        lower boundary of the sample interval (inclusive)
    high: float
        upper boundary of the sample interval (exclusive)

    Returns
    -------
    float:
        a random value sampled according to a uniform distribution
        between *low* and *high*

    See also
    --------
    numpy.random.uniform

    """
    return np.random.uniform(low, high)

random_uniform.__USER_FUNCTION__ = ['Functions Available for Models']

def random_integers(low: int, high: int) -> int:
    """Return a random integer value drawn from a discrete uniform
    distribution between *low* and *high* (both inclusive).

    Example
    -------
    In a prototype definition::

      age: random_integers(min_age, max_age)

    Parameters
    ----------
    low: int
        lower boundary of the sample interval (inclusive)
    high: int
        upper boundary of the sample interval (inclusive)

    Returns
    -------
    int:
        a random integer value sampled according to a discrete uniform
        distribution between *low* and *high*

    See also
    --------
    numpy.random.random_integers
    """
    return np.random.random_integers(low, high)

random_integers.__USER_FUNCTION__ = ['Functions Available for Models']

def random_exponential(scale: float) -> float:
    """Return a random value drawn from an exponential distribution of
    rate 1/*scale* (thus of mean *scale*).

    Example
    -------
    In a prototype definition::

      time_to_live: random_exponential(mean_duration)

    Parameters
    ----------
    scale: float
        the scale parameter of the distribution, i.e. the inverse of the rate

    Returns
    -------
    float:
        a random value sampled according to an exponential
        distribution of rate 1/*scale*

    See also
    --------
    numpy.random.exponential

    """
    return np.random.exponential(scale)

random_exponential.__USER_FUNCTION__ = ['Functions Available for Models']

def random_beta(a: float, b: float) -> float:
    """Return a random value drawn from a beta distribution of parameters
    *a* and *b*.

    Example
    -------
    In a prototype definition::

      age: 'random_beta(2, 5) * age_max'

    Parameters
    ----------
    a: float, positive (>0)
        first shape parameter of the beta distribution
    b: float, positive (>0)
        second shape parameter of the beta distribution

    Returns
    -------
    float:
        a random value sampled according to a beta distribution of
        parameters *a* and *b*

    See also
    --------
    numpy.random.beta
    """
    return np.random.beta(a, b)

random_beta.__USER_FUNCTION__ = ['Functions Available for Models']

def random_gamma(shape: float, scale: float) -> float:
    """Return a random value drawn from a gamma distribution of parameters
    *shape* and *scale*.

    Example
    -------
    In a prototype definition::

      age: 'random_gamma(3, 2)'

    Parameters
    ----------
    shape: float, positive (>0)
        the shape of the gamma distribution
    scale: float, positive (>0)
        the scale of the gamma distribution

    Returns
    -------
    float:
        a random value sampled according to a gamma distribution of
        parameters *shape* and *scale*

    See also
    --------
    numpy.random.gamma
    """
    return np.random.gamma(shape, scale)

random_gamma.__USER_FUNCTION__ = ['Functions Available for Models']


def random_normal(mn: float, sd: float) -> float:
    """Return a random value drawn from a normal distribution of mean *mn*
    and standard deviation *sd*.

    Example
    -------
    In a prototype definition::

      age: 'random_normal(100, 5)'

    Parameters
    ----------
    mn: float
        the mean of the normal distribution
    sd: float, positive (>=0)
        the standard deviation of the normal distribution

    Returns
    -------
    float:
        a random value sampled according to a normal distribution of
        mean *mn* and standard deviation *sd*

    See also
    --------
    numpy.random.normal

    """
    return np.random.normal(mn, sd)

random_normal.__USER_FUNCTION__ = ['Functions Available for Models']


def random_poisson(lam: float) -> int:
    """Return an integer random value drawn from a Poisson distribution of
    mean *lam*.

    Example
    -------
    In an action, e.g. here when computing how many newborn
    individuals will be produced::

      - from: Gestating
        to: NonGestating
        on_cross:
          - produce_offspring: newborn
            amount: 'random_poisson(1.05)'

    Parameters
    ----------
    lam: float
        the mean of the Poisson distribution

    Returns
    -------
    int:
        a random integer value sampled according to a Poisson distribution of
        mean *lam*

    See also
    --------
    numpy.random.poisson

    """
    return np.random.poisson(lam)

random_poisson.__USER_FUNCTION__ = ['Functions Available for Models']



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Tools aimed at handling graphs for the state machines
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   collections               import OrderedDict

from   emulsion.tools.state      import EmulsionEnum


EdgeTypes = EmulsionEnum('EdgeTypes', ['TRANSITION', 'PRODUCTION'], module=__name__)
EdgeTypes.linestyle = 'solid'
EdgeTypes.PRODUCTION.linestyle = 'dashed'



class MultiDiGraph:
    """An oriented multigraph (two nodes can be linked by several
    edges). Replacement for MultiDiGraph in networkx with a
    reproducible order in accessing edges and nodes.

    """
    def __init__(self, **attributes):
        """Create an instance of MultiDiGraph. If keywords arguments
        are specified, they are considered attributes of the whole
        graph.

        """
        self.graph = dict(**attributes)
        self.node = OrderedDict()
        self.edge = OrderedDict()
        self._edge_counter = OrderedDict()

    def add_node(self, node_id, **attributes):
        """Add the specified node to the graph. Node attributes can be
        specified. If the node is already present, updated the
        attributes.

        """
        if node_id not in self.node:
            self.node[node_id] = dict()
        self.node[node_id].update(attributes)


    def add_edge(self, from_id, to_id, type_id=EdgeTypes.TRANSITION,
                 key=None, **attributes):
        """Add the specified edge to the graph. If nodes are not
        already created, they are automatically added. Edge attributes
        can be specified. Since this graph allows multiple edges
        between pairs of nodes, two calls to this method lead to two
        distinct edges (even if the attributes are the same), unless a
        key is specified. The key is used to identify edges between a
        given pair of nodes. By default, keys are consecutive
        integers.

        """
        self.add_node(from_id)
        self.add_node(to_id)
        if from_id not in self.edge:
            self.edge[from_id] = OrderedDict()
            self._edge_counter[from_id] = OrderedDict()
        if to_id not in self.edge[from_id]:
            self.edge[from_id][to_id] = OrderedDict()
            self._edge_counter[from_id][to_id] = 0
        if key is None:
            edge_key = self._edge_counter[from_id][to_id]
            self._edge_counter[from_id][to_id] += 1
        else:
            edge_key = key
        attributes['type_id'] = type_id
        self.edge[from_id][to_id][edge_key] = dict(**attributes)

    def edges(self):
        """Return a list of edge tuples."""
        return [(from_id, to_id)
                for from_id in self.edge
                for to_id in self.edge[from_id]]

    def edges_from(self, from_id, type_id=EdgeTypes.TRANSITION):
        """Return a list of tuples (to_id, attributes) corresponding
        to all edges going out of the `from_id` node.

        """
        if from_id not in self.edge:
            return []
        return [(state, attribs)
                for state in self.edge[from_id]
                for attribs in self.edge[from_id][state].values()
                if (type_id is None) or (attribs['type_id'] == type_id)]



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Classes and functions for process management in MultiProcessCompartments.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   abc                       import abstractmethod
import numpy                     as np

from   emulsion.tools.misc       import rates_to_probabilities, aggregate_probabilities, count_population

class AbstractProcess(object):
    """An AbstractProcess is aimed at controlling a specific activity
    in a compartment, and is identified by its name.

    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Process "{}"'.format(self.name)

    __str__ = __repr__

    @abstractmethod
    def evolve(self):
        """Define the actions that the process must perform."""
        pass


class MethodProcess(AbstractProcess):
    """A MethodProcess is aimed at running a specific method (and
    possibly any function or even any callable object).

    """
    def __init__(self, name, method, lparams=[], dparams={}):
        super().__init__(name)
        self.method = method
        self.lparams = lparams
        self.dparams = dparams

    def evolve(self):
        """Define the actions that the process must perform. In a
        MethodProcess, those actions consist in running the method of
        the target compartment.

        """
        self.method(*self.lparams, **self.dparams)


class StateMachineProcess(AbstractProcess):
    """A StateMachineProcess is aimed at running a specific state machine
    within the agent (not within a grouping).

    """
    def __init__(self, name, agent, state_machine):
        super().__init__(name)
        self.agent = agent
        self.state_machine = state_machine

    def evolve(self):
        """Define the actions that the process must perform. In a
        StateMachineProcess, those actions consist in 1) executing the
        transitions of the state machine to change the agent's states
        and 2) computing productions and transmit them to the upper
        level of the agent.

        """
        self.evolve_transitions()
        self.agent.upper_level().new_agents += self.evolve_productions()

    def evolve_transitions(self):
        # retrieve the name of the statevar/attribute where the
        # current state is stored
        statevar = self.state_machine.machine_name
        # retrieve the value of the current state
        #current_state = self.get_information(statevar).name
        current_state = self.agent.statevars[statevar].name
        # perform actions associated to the current state
        self.agent.do_state_actions('on_stay', self.state_machine,
                                    current_state, population=1)
        # retrieve all possible transitions from this state
        transitions = self.agent.next_states_from(current_state,
                                                  self.state_machine)
        # skip this machine if no available transitions
        if transitions:
            states, flux, values, _, actions = zip(*transitions)
            total_value = sum(values)
            states = states + (current_state,)
            actions = actions + ([], )
            available_flux = set(flux)
            if 'amount' in available_flux or 'amount-all-but' in available_flux:
            # handle amounts
                # compute proper values (bounded by 0/1) and when
                # needed, invert 'amount-all-but' values
                values = [max(min(1-v, 1), 0) if f == 'amount-all-but'\
                            else max(min(v, 1), 0)
                          for (f, v) in zip(flux, values)]
                # recompute total value
                total_value = sum(values)
                # normalize to have probabilities
                if total_value == 0:
                    values = (0,)*len(values) + (1,)
                else:
                    values = tuple(v / total_value
                                   for v in values) + (1- total_value,)
            elif 'proba' in available_flux:
                # handle probabilities
                values = aggregate_probabilities(values,
                                                 self.agent.model.delta_t)
                values += (1 - sum(values),)
            else:
                # transform rates into probabilities
                values = rates_to_probabilities(total_value, values,
                                                delta_t=self.agent.model.delta_t)
            index = np.nonzero(np.random.multinomial(1, values))[0][0]
            next_state = states[index]
            next_action = actions[index]
            # condition on the index, so as to allow reflexive edges !
            if index != len(values)-1:
            # if next_state != current_state:
                self.agent.do_state_actions('on_exit', self.state_machine,
                                            current_state, population=1)
                # self.agent.set_information(statevar,
                #                            self.state_machine.states[next_state])
                self.agent.change_state(statevar,
                                        self.state_machine.states[next_state])
                # self.agent.do_edge_actions('on_cross', state_machine,
                #                      current_state, next_state, population=1)
                if next_action:
                    self.agent.do_edge_actions(actions=next_action, population=1)
                self.agent.do_state_actions('on_enter', self.state_machine,
                                            next_state, population=1)

    def evolve_productions(self):
        # init empty list for all changes to perform
        future = []
        # retrieve the name of the statevar/attribute where the
        # current state is stored
        statevar = self.state_machine.machine_name
        # retrieve the value of the current state
        current_state = self.agent.statevars[statevar].name
        # retrieve all possible productions from this state
        productions = self.agent.production_from(current_state,
                                                 self.state_machine)
        # skip this machine if no available transitions
        if productions:
            ### HERE WE ASSUME THAT AN AGENT CAN PRODUCE SEVERAL OTHER
            ### AGENTS SIMULTANEOUSLY (OTHERWISE USE CONDITIONS)
            ### REWRITE TRANSITIONS TO HAVE DISJOINT SUB-POPULATIONS
            for target_state, flux, values, ref_pop, proto in productions:
                pop_size = count_population(ref_pop)
                amount = self.agent._compute_production(values, flux, pop_size, self.agent.stochastic)
                if amount > 0:
                    future.append((target_state, amount, proto))
        return future




# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""

"""
from   pathlib               import Path
import csv
import dateutil.parser       as     dup
import numpy                 as     np
from emulsion.agent.managers import MetapopProcessManager

DATA_FILE = 'moves.csv'

#===============================================================
# CLASS Metapopulation (LEVEL 'metapop')
#===============================================================
class Metapopulation(MetapopProcessManager):
    """
    level of the metapop.

    """
    def initialize_level(self, **others):
        """Initialize an instance of Metapopulation.
        Additional initialization parameters can be introduced here if needed.
        """
        # read a CSV data file for moves:
        # date of movement, source herd, destination herd, age group, quantity

        # and restructure it according to origin_date and delta_t:
        # {step: {source_id: [(dest_id, age_group, qty), ...],
        #         ...},
        #  ...}
        origin = self.model.origin_date
        step_duration = self.model.step_duration
        moves = {}
        with open(Path(self.model.input_dir, DATA_FILE)) as csvfile:
            # read the CSV file
            csvreader = csv.DictReader(csvfile, delimiter=',')
            for row in csvreader:
                day = dup.parse(row['date'])
                if day < origin:
                    # ignore dates before origin_date
                    continue
                # convert dates into simulation steps
                step = (day - origin) // step_duration
                # group information by step and source herd
                if step not in moves:
                    moves[step] = {}
                src, dest, qty = int(row['source']), int(row['dest']), int(row['qty'])
                if src not in moves[step]:
                    moves[step][src] = []
                moves[step][src].append([dest, row['age'], qty])
        self.moves = moves

    #----------------------------------------------------------------
    # Processes
    #----------------------------------------------------------------
    def exchange_animals(self):
        """

        => INDICATE HERE HOW TO PERFORM PROCESS exchange_animals.
        """
        if self.statevars.step in self.moves:
            herds = self.get_populations()
            for source in self.moves[self.statevars.step]:
                for dest, age, qty in self.moves[self.statevars.step][source]:
                    # neither source/dest in simulated herds
                    if (source not in herds) and (dest not in herds):
                        # print('ignoring movement from {} to {} (outside the metapopulation)'.format(source, dest))
                        continue
                    # source not in simulated herds: create animal from prototype
                    if source not in herds:
                        # print('movement to {} coming from outside the metapopulation'.format(dest))

                        # retrieve prototype definition from the model
                        prototype = self.model.get_prototype(name='imported_movement', level='animals')
                        # change age group to comply with movement
                        prototype['age_group'] = self.get_model_value(age)
                        animals = [herds[dest].new_atom(custom_prototype=prototype)
                                   for _ in range(qty)]
                        # print('importing', animals)
                    else:
                        # find convenient animals
                        candidates = herds[source].select_atoms('age_group', age, process='aging')
                        # try to move the appropriate quantity
                        nb = min(len(candidates), qty)
                        animals = np.random.choice(candidates, nb, replace=False)
                        herds[source].remove_atoms(animals)
                        # print('removing', animals, 'from', source)
                        # update variable 'sold' in source herd
                        herds[source].statevars.sold += len(animals)
                    if dest not in herds:
                        pass
                        # print('movement from {} going outside the metapopulation'.format(source))
                    else:
                        herds[dest].add_atoms(animals)
                        # update variable 'purchased' in dest herd
                        herds[dest].statevars.purchased += len(animals)
                        # print('adding', animals, 'to', dest)



"""
.. module:: emulsion.agent.managers.functions

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from   collections               import OrderedDict

import numpy                     as np

from   emulsion.agent.exceptions import LevelException
from   emulsion.agent.views      import AdaptiveView
from   emulsion.tools.misc       import load_class, select_random,\
    add_all_test_properties, AGENTS


from   emulsion.agent.managers.group_manager  import  GroupManager
from   emulsion.agent.managers.abstract_process_manager  import  AbstractProcessManager


#  __  __       _ _   _ _____
# |  \/  |     | | | (_)  __ \
# | \  / |_   _| | |_ _| |__) | __ ___   ___ ___  ___ ___
# | |\/| | | | | | __| |  ___/ '__/ _ \ / __/ _ \/ __/ __|
# | |  | | |_| | | |_| | |   | | | (_) | (_|  __/\__ \__ \
# |_|  |_|\__,_|_|\__|_|_|   |_|  \___/ \___\___||___/___/
#  __  __
# |  \/  |
# | \  / | __ _ _ __   __ _  __ _  ___ _ __
# | |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |  | | (_| | | | | (_| | (_| |  __/ |
# |_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                            __/ |
#                           |___/

class MultiProcessManager(AbstractProcessManager):
    """A MultiProcessManager is aimed handling several independent
    StructuredViews at the same time, together with a
    SimpleView containing all the atom units. It can
    automatically build compartments for:
    - state machines associated with a specific state variable or attribute
    - specific state variables or attributes with a limited number of
    values, such as booleans or enumerations

    """
    def __init__(self, model=None, level=None, **others):
        view_class, options = load_class(
            **model.levels[level]['super']['master']
        )
        master = view_class(keys='MASTER', host=self, recursive=True, **options)
        super().__init__(model=model, master=master, level=level, **others)


    def apply_initial_conditions(self):
        """Initialize level with initial conditions specified in the model.

        As this agent is aimed at managing individuals in the
        sub-levels, only 'prototypes' are taken into account to
        initialize the sub-levels.

        """
        if self.level in self.model.initial_conditions:
            conds = self.model.initial_conditions[self.level]
            for protos, qty, probas in conds['prototypes']:
                proba_values = [self.get_model_value(p) for p in probas]
                total_value = sum(proba_values)
                assert(0 <= total_value <= 1)
                ## if sum of probas < 1 add complement (even if
                ## proba_values same size as protos) to ensure that
                ## e.g. [0.1, 0.1] with amount 100 gives in average
                ## [10, 10] individuals
                if total_value < 1:
                    proba_values += [1 - total_value]
                amount = int(self.get_model_value(qty))
                qty_by_proto = np.random.multinomial(amount, proba_values)
                # truncate qty_by_proto if size > nb of protos actually used
                qty_by_proto = qty_by_proto[:len(protos)]
                to_add = []
                for proto, nb in zip(protos, qty_by_proto):
                    to_add += [ self.new_atom(prototype=proto, execute_actions=True)
                                for _ in range(nb)
                    ]
                self.add_atoms(to_add, init=True)

    def add_host(self, host):
        """Add the specified host to the current Multiprocessmanager, associated
        with the specified key.

        """
        if self._host is None:
            self._host = OrderedDict()
        self._host[host.keys] = host
        self.simulation = host.simulation

    def add_new_population(self, process_name, population):
        """Create new atoms with the population information"""
        new_atoms = []
        for target_state, amount, proto in population:
            val = self.get_model_value(target_state)
            var = val.state_machine
            sublevel = self.get_default_sublevel()
            prototype = self.model.get_prototype(sublevel, proto)
            prototype[var.machine_name] = val
            to_add = [ self.new_atom(custom_prototype=prototype, execute_actions=True)
                       for _ in range(int(amount)) ]
            # print(var, '=', val)
            # for atom in to_add:
            #     atom.change_state(var, val)
            new_atoms += to_add
        self.add_atoms(new_atoms)

    def add_compart_process(self,
                            process_name,
                            key_variables,
                            compart_manager=(GroupManager, {}),
                            machine_name=None,
                            allowed_values=None,
                            compart_class=(AdaptiveView, {})):
        super().add_compart_process(process_name, key_variables,
                                    compart_manager=compart_manager,
                                    machine_name=machine_name,
                                    allowed_values=allowed_values,
                                    compart_class=compart_class)


    def select_atoms(self, variable=None, state=None, value=None, process=None):
        """Return a list of atoms selected by specific *value* or *state* of a
        *variable*.

        Parameters
        ----------
        variable: str
            the variable to be compared to the *value* - if ``None``,
            all atoms are selected.
        state: str
            a name of the state used for the selection (instead of *value*)
        value: object
            the value to use for selection - if ``None``, the value is
            replaced with the model state specified in parameter
            *state*
        process: str
            name of a process associated to a specific grouping based
            on the *variable* (to accelerate search)

        Returns
        -------
        list:
            the list of matching agents in the sublevel
        """
        if variable is None:
            return list(self['MASTER'])
        if state is not None:
            value = self.model.get_value(state)
        if process is not None:
            if (value,) in self[process]._content:
                return list(self[process][(value,)]._content)
        return [agent for agent in self['MASTER']
                if agent.get_information(variable) == value]


    def get_group_atoms(self, process_name, group_name):
        """Return all atoms which belong to the the specified group
        name (state names). (*group_name* may be a subset of the grouping key)

        """
        complete_key = set(self.get_model_value(state_name)
                           for state_name in group_name)
        groups = self[process_name]
        value = []
        for key, compart in groups._content.items():
            if complete_key <= set(key):
                value += compart.get_content()[AGENTS]
        return value



    def get_agent_class_for_sublevel(self, sublevel):
        """Return the agent class in charge of representing the specified
        sublevel.

        """
        if sublevel not in self.model.levels:
            raise LevelException('not found', sublevel)
        if 'contains' not in self.model.levels[self.level]:
            raise LevelException('not linked with %s' % (self.level), sublevel)
        cl, _ = self.model.get_agent_class_for_level(sublevel)
        return cl

    def get_default_sublevel(self):
        """Return by default the first sublevel contained in this level, if
        any. If this level contains no sublevels, raise a
        LevelException.

        """
        if 'contains' not in self.model.levels[self.level]:
            raise LevelException('not specified for atom creation', '')
        return self.model.levels[self.level]['contains'][0]

    def new_atom(self, sublevel=None, prototype=None, custom_prototype=None,
                 execute_actions=False, **args):
        """Instantiate a new atom for the specified sublevel, with the
        specified arguments. If the sublevel is not specified, the
        first one from the `contains` list is taken. If the name of a
        prototype is provided, it is applied to the new agent (using
        the `execute_actions` parameter).

        """
        # print('Entering new_atom for', self, 'with prototype', prototype)
        if sublevel is None:
            sublevel = self.get_default_sublevel()
        atom_class = self.get_agent_class_for_sublevel(sublevel)
        args.update(model=self.model, simu_id=self.statevars.simu_id,
                    level=sublevel, step=self.statevars.step)
        if prototype is not None:
            args.update(prototype=prototype, execute_actions=execute_actions)
        elif custom_prototype is not None:
            args.update(custom_prototype=custom_prototype, execute_actions=execute_actions)
        new_atom = atom_class(**args)
        # initialize MASTER host
        new_atom.add_host(self['MASTER'])
        add_all_test_properties(new_atom)
        new_atom.apply_initial_prototype(name=prototype, prototype=custom_prototype,
                                         execute_actions=execute_actions)
        return new_atom

    def add_atoms(self, atom_set, init=False, level=None):
        """Add the specified set of atoms to the current
        MultiProcessManager. Atoms are especially added
        automatically to each of the compartment managers.  If `init`
        is True, the compartment managers counts the initial value of
        the populations in each compartment.

        """
        self['MASTER'].add(atom_set)
        self.statevars.population = len(self['MASTER']._content)
        # update the model of atoms
        if level is None:
            if 'contains' in self.model.levels[self.level]:
                level = self.model.levels[self.level]['contains'][0]
        for atom in atom_set:
            atom.model = self.model
            atom.level = level
        # check if any initialization action has to be performed
        for machine in self.init_machines:
            agents_to_init = OrderedDict()
            for atom in atom_set:
                state = atom.statevars[machine]
                if state in self.model.init_actions[machine]:
                    if state not in agents_to_init:
                        agents_to_init[state] = []
                    agents_to_init[state].append(atom)
            for state, atoms in agents_to_init.items():
                for action in self.model.init_actions[machine][state]:
                    # print(f'Executing INIT action {action} for state {state.name}')
                    action.execute_action(self['MASTER'], agents=atoms)

        # add atoms to appropriate compartments and make them
        # consistent
        for name, comp in self._content.items():
            if name not in self.no_compart:
                default_key = tuple(None for _ in comp.keys)
                comp[default_key].add(atom_set)
                self.make_consistent(comp)
                if init:
                    comp.update_counts()

    def make_all_consistent(self):
        """Check all compartments to ensure their consistency"""
        for name, comp in self._content.items():
            if name not in self.no_compart:
                self.make_consistent(comp)
                comp.update_counts()

    def make_consistent(self, compartment):
        """Make the specified dict compartment check and handle the
        consistency of its own sub-compartments.

        """
        for comp in compartment:
            comp.check_consistency()
        compartment.handle_notifications()

    def remove(self, agents_or_population):
        assert(agents_or_population[0] == 'agents')
        self.remove_atoms(agents_or_population[1])

    def remove_atoms(self, atom_set):
        """Remove the specified atoms from the current
        MultiProcessManager. Atoms are removed from each of the
        compartment managers (including the 'MASTER' set).

        """
        for atom in list(atom_set):
            for host in list(atom._host.values()):
                host.remove([atom])
        self.statevars.population = len(self['MASTER']._content)

    def select_randomly(self, proba=0, amount=None, process=None):
        """Select randomly chosen atoms from this ProcessManager. `proba` can
        be either a probability or a dictionary. In that case, the
        `process` parameter indicates the name of the process grouping
        which drives the probabilities, and the keys must be those of
        the grouping. Selected atoms are removed and returned by the
        method.

        """
        if self.population <= 0:
            return []
        if process is None:
            if amount is None:
                amount = np.random.binomial(len(self['MASTER']), proba)
            selection = select_random(self['MASTER'], amount)
        else:
            selection = []
            if amount is None:
                for key, compart in self[process].items():
                    if key in proba:
                        selection += select_random(
                            compart, np.random.binomial(len(compart),
                                                        proba[key]))
            else:
                return []       #  inconsistent call
        return selection

    def remove_randomly(self, proba=0, amount=None, process=None):
        """Remove randomly chosen atoms from this ProcessManager. `proba` can
        be either a probability or a dictionary. In that case, the
        `process` parameter indicates the name of the process grouping
        which drives the probabilities, and the keys must be those of
        the grouping. Selected atoms are removed and returned by the
        method.

        """
        to_remove = self.select_randomly(amount=amount, proba=proba,
                                         process=process)
        self.remove_atoms(to_remove)
        # print(to_remove)
        return to_remove



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Plotting tools... to be improved!
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from   pathlib                      import Path

import sys
import pandas                       as pd
import numpy                        as np

# from   scipy.misc                   import imread
from   PIL                          import Image
from   xml.etree                    import ElementTree
from   bokeh.plotting               import figure, output_file, show
from   bokeh.models                 import Legend, ColumnDataSource, Range1d
from   bokeh.models.tools           import HoverTool, CrosshairTool
from   bokeh.models.glyphs          import ImageURL
from   bokeh.models.widgets         import Div
from   bokeh.layouts                import gridplot, column, widgetbox
from   bokeh.palettes               import all_palettes


def build_state_plot(counts, cols, machine, model, y='quantity', group='state',
                     ylab='Number of individuals'):
    """Return a bokeh figure based on the representation of the states in
    the *counts* dataframe, with the associated colors.

    Args:
        counts: a Pandas dataframe containing the values to plot
        cols (dict): dictionary which maps state/variable names to colors
        machine: the state machine to which the states are related
        model: the Emulsion model related to this plot
        y (str): the name of the field containing y values in the dataframe
        group (str): either 'state' (default) or 'variables', name of
          the field containing each legend items.

    Returns:
        A bokeh gridplot (reduced to one figure a the herd level, one
        figure per herd at metapopulation level).

    """
    WIDTH, HEIGHT = 800, 600
    NB_SIMU = 1 + counts.simu_id.max()
    alpha = max(1 /np.sqrt(NB_SIMU), 0.05)
    all_plots = []
    ncols = 1
    leg_pos, leg_orient = 'right', 'vertical'
    metapop = 'population_id' in counts.columns
    if metapop:
        max_pop_id = 1 + max(counts.population_id)
        ncols = 2
        WIDTH = 1200
    else:
        max_pop_id = 1
    for pop_id in range(max_pop_id):
        if metapop:
            df = counts[counts.population_id == pop_id]
        else:
            df = counts
        time_unit = model.time_unit
        origin = model.origin_date
        dur = model.step_duration
        df[time_unit] = pd.Series(model.delta_t * df['step'], index=df.index)
        # df['date'] = pd.Series(pd.to_datetime(origin + dur * df['step']), index=df.index)
        plot = figure(plot_width=WIDTH // ncols, plot_height=HEIGHT // ncols)
        legend_items = []
        for state in cols:
            state_items = []
            subdf = df[df[group] == state]
            if NB_SIMU <= 50:   # don't draw points if too many !
                ci = plot.circle(x=time_unit, y='quantity', source=ColumnDataSource(subdf),
                                 size=2, color=cols[state], alpha=alpha)
                state_items.append(ci)
            for simu_id in range(1+max(df.simu_id)):
                src = ColumnDataSource(subdf[subdf.simu_id == simu_id])
                pl = plot.line(x=time_unit, y='quantity', source=src, color=cols[state], alpha=alpha, line_width=2)
                state_items.append(pl)
            legend_items.append((state, state_items))
        plot.title.text = 'Evolution of %s (%s)' % (machine.replace('_', ' '), model.model_name)
        if metapop:
            plot.title.text += ' - population ID: {}'.format(pop_id)
        plot.xaxis.axis_label = 'Time (%s)' % (time_unit,)
        plot.yaxis.axis_label = ylab
        ylim = max(1, df['quantity'].max()*1.1)
        plot.y_range = Range1d(0, ylim)
        hover = HoverTool()
        hover.tooltips=[
            (group.capitalize(), '@%s' % (group,)),
            ('Quantity', '@quantity'),
            (time_unit.capitalize(), '@%s'%(time_unit,)),
            ('Step', '@step'),
            ('Simulation ID', '@simu_id'),
            # ('Population ID', pop_id)
        ]
        plot.add_tools(hover)
        plot.add_tools(CrosshairTool())
        legend = Legend(items=legend_items, location=(0, 0))
        legend.click_policy = 'hide'
        legend.orientation = leg_orient
        plot.add_layout(legend, leg_pos)
        all_plots.append(plot)
    return gridplot(all_plots, ncols=ncols, toolbar_location="above")


def build_machine_plot(machine_name, model, params):
    """Return a bokeh figure based on the representation of the state
    machine, with the associated colors.

    Args:
        machine: the state machine to which the states are related

    Returns:
        A bokeh gridplot.

    """
    # WIDTH, HEIGHT = 800, 600
    machine_url = model.model_name + '_' + machine_name + '_machine.' + params.img_format
    machine_path = Path(params.figure_dir, machine_url)
    try:
        # img = imread(machine_path)
        # height, width = img.shape[0:2]
        if params.img_format == 'svg':
            img = ElementTree.parse(machine_path)
            root = img.getroot()
            # retrieve dimensions (in pts)
            width = int(''.join(i for i in root.attrib['width'] if i.isdigit()))
            height = int(''.join(i for i in root.attrib['height'] if i.isdigit())) + 50
        else:
            img = Image.open(machine_path)
            width, height = img.size
    except:
        print("WARNING: Could not retrieve image dimensions")
        width, height = 800, 400

    div = Div(text='''<h1>State machine for {}</h1>
    <P><IMG SRC="{}"></P>
    <H2>Simulation outputs</H2>
    '''.format(machine_name, machine_url),
              width=width, height=height+100)

    return gridplot([widgetbox(div)], ncols=1)

def plot_outputs(params):
    """Read outputs from previous runs and plot the corresponding
    figures. In the *params* dictionary, `output_dir` is expected to
    contain a `counts.csv` file; `figure_dir` is where the plot is
    saved.

    """
    countpath = Path(params.output_dir, 'counts.csv')
    if not countpath.exists():
        print('ERROR, output file not found: %s' % (countpath, ))
        sys.exit(-1)
    model = params.model
    df = pd.read_csv(countpath)
    vars = ['simu_id', 'step', 'level', 'agent_id']
    if 'population_id' in df.columns:
        vars.append('population_id')
    vals = []
    all_figs = []
    figpath = Path(params.figure_dir, model.model_name + '.html')
    output_file(figpath, title='Emulsion Plot: %s' % (model.model_name,))
    for sm_name, state_machine in model.state_machines.items():
        if params.view_machines:
            all_figs.append(build_machine_plot(sm_name, model, params))
        col_dict = state_machine.state_colors
        # col_dict.update(population="black")
        states, cols = zip(*sorted(col_dict.items()))
        vals += [state.name for state in state_machine.states]
        df2 = pd.melt(df, id_vars=vars, value_vars=states,
                      var_name='state', value_name='quantity')
        fig = build_state_plot(df2, col_dict, sm_name, model)
        all_figs.append(fig)
        # figpath = Path(params.figure_dir,
        #                model.model_name + '_' + sm_name + '.png')
        # plot = build_state_plot(df2, cols, sm_name, model.model_name)
        # plot.save(str(figpath))
        # print('Saved figure %s' % (figpath, ))
    extras = [variable for variable in df.columns if variable not in vars + vals]
    if extras:
        df2 = pd.melt(df, id_vars=vars, value_vars=extras, value_name='quantity')
        idx = [(256 // len(extras))*i for i in range(len(extras))]
        cols=dict(zip(extras, [all_palettes['Plasma'][256][i] for i in idx]))
        fig = build_state_plot(df2, cols, 'Additional variables', model,
                               y='value', group='variable', ylab='Value')
        # plot = build_state_plot(df2, [], 'Additional variables',
        #                         model.model_name, y='value', group='variable')
        # plot.save(str(figpath))
        # print('Saved figure %s' % (figpath, ))
        all_figs.append(fig)
    # print(all_figs)
    print('Outputs plot in file: %s' % (figpath,))
    show(column(all_figs))



"""
.. module:: emulsion.agent.core.emulsion_agent

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

Part of this code was adapted from the PADAWAN framework (S. Picault,
Univ. Lille).
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

import numpy                      as np


from   emulsion.tools.graph       import EdgeTypes
from   emulsion.model             import EDGE_KEYWORDS
from   emulsion.tools.misc        import rates_to_probabilities,\
    aggregate_probabilities, probabilities_to_rates, count_population, AGENTS

from   emulsion.agent.core.abstract_agent  import AbstractAgent


#  ______                 _     _                                      _
# |  ____|               | |   (_)               /\                   | |
# | |__   _ __ ___  _   _| |___ _  ___  _ __    /  \   __ _  ___ _ __ | |_
# |  __| | '_ ` _ \| | | | / __| |/ _ \| '_ \  / /\ \ / _` |/ _ \ '_ \| __|
# | |____| | | | | | |_| | \__ \ | (_) | | | |/ ____ \ (_| |  __/ | | | |_
# |______|_| |_| |_|\__,_|_|___/_|\___/|_| |_/_/    \_\__, |\___|_| |_|\__|
#                                                      __/ |
#                                                     |___/


class EmulsionAgent(AbstractAgent):
    """The EmulsionAgent is the base class for multi-level
    epidemiologic agents. An EmulsionAgent can represent any entity
    involved in the epidemiologic model (individual, compartment,
    herd, [meta]population, etc.).

    Each agent contains an exchange box (in & out) which is composed
    of a list of messages. A message is generally a dictionary of
    Source/Destination (In/Out) and content of exchange with
    Source/Destination agent.

    """
    def __init__(self, model=None, level=None, name=None, host=None,
                 simulation=None, **others):
        """Initialize the unit with a health state and a name."""
        super().__init__(**others)
        # self.statevars.health_state = health_state
        self.model = model
        self.level = level
        self._name = name
        self._host = host
        self.simulation = simulation
        if 'step' not in self.statevars:
            self.statevars.step = 0
        self.statevars.creation_step = self.statevars.step
        # exchange inbox/outbox
        self._inbox = []
        self._outbox = []

    def upper_level(self, init=True):
        """Return the 'upper level' for this agent, i.e. the first host with
        a not-None level attribute.

        TAG: USER
        """
        if self.level is not None and not init:
            return self
        return self.get_host().upper_level(init=False)

    def top_level(self):
        """Return the highest level used in the simulation"""
        if self.get_host() is None:
            return self
        return self.upper_level().top_level()

    def log_path(self):
        """Return the path to the log file for the current simulation"""
        return self.top_level().simulation.log_path()

    @property
    def name(self):
        """Return the name of the unit. If no name was provided during
        instantiation,  the class name is returned.

        TAG: USER
        """
        return self._name if self._name is not None\
            else super().__repr__() # self.__class__.__name__

    def __repr__(self):
        return self.name


    def evolve(self, machine=None):
        """This method is aimed at defining what has systematically to
        be done in the unit at each time step (e.g. age change...). It
        has to be overriden if needed in subclasses.

        """
        self.statevars.step += 1

    def get_host(self, key=None):
        """Return the host of the current unit.

        TAG: USER
        """
        return self._host

    @abc.abstractmethod
    def get_content(self):
        """Return either the population (number) of the current unit,
        or the list of agents contained in the current unit. The
        output is a tuple with either ('population', qty) or ('agents', list).

        """
        pass

    def do_state_actions(self, event, state_machine, state_name,
                         population=None, agents=None, **_):
        """Perform the actions associated to the current state. If the
        unit is a ViewAgent, actions are actually performed by each
        unit of the specified agents list, in turn. Otherwise, the
        actions are performed according to the population, which is
        expected to be a number.

        """
        # if actions are actually associated to the current state of
        # the state machine...
           # ... and to the 'event' (enter/exit/stay)
        if state_name in state_machine.state_actions\
           and event in state_machine.state_actions[state_name]:
            # print(f'Doing state actions {event} for {state_machine}\n\tfor {state_name}')
            # retrieve the list of actions
            l_actions = state_machine.state_actions[state_name][event]
            # ask the current unit to perform the actions with the
            # specified population
            # print(f'\t{l_actions}')
            for action in l_actions:
                action.execute_action(self, population=population, agents=agents)

    def do_edge_actions(self, actions=None, population=None, agents=None):
        """Perform the actions associated to a transition between
        states. If the unit is a ViewCompartment, actions are actually
        performed by each unit of the specified agents list, in
        turn. Otherwise, the actions are performed according to the
        population, which is expected to be a number.

        """
        # # if actions are actually associated to the current edge...
        #    # ... and to the 'event' (cross)
        # if from_ in state_machine.edge_actions\
        #    and to_ in state_machine.edge_actions[from_]\
        #    and event in state_machine.edge_actions[from_][to_]:
        #     # retrieve the list of actions
        #     l_actions = state_machine.edge_actions[from_][to_][event]
        #     # ask the current unit to perform the actions with the
        #     # specified population
        for action in actions:
#            print(action)
            action.execute_action(self, population=population, agents=agents)


    def next_states_from(self, initial_state, state_machine):
        """Return a list of tuples composed of:

        - each of the possible states reachable from the specified
        initial state (some depending on a condition)

        - the transition rate, probability or amount to each state in
        the specified state_machine (possible clauses: 'rate',
        'proba', 'amount', 'amount-all-but' with the actual value)

        - a tuple indicating who can cross the edge, depending on
        conditions (either ('population', qty) or ('agents', list))

        - the list of actions to perform when crossing the edge (if
        any).

        The conditions, rates and probabilities may depend on the state
        variables or properties of the current unit.

        """
        result = []
        # remove unfulfilled 'when' clauses if any
        for (state, props) in state_machine.graph.edges_from(initial_state,
                                                             type_id=EdgeTypes.TRANSITION):
            if 'when' in props:
                when = state_machine.get_value(props['when'])
                fulfilled = when(self) if callable(when) else when
                if not fulfilled:
                    continue
            cond_result = self.get_content()
            # if any condition, evaluate it
            if 'cond' in props:
                cond = state_machine.get_value(props['cond'])
                # compute the content tuple (either ('population', qty)
                # or ('agents', lsit)) which fulfils the condition
                cond_result = self.evaluate_condition(cond)
            # only edges with condition fulfilled are taken into account
            if cond_result:
                # print(cond_result)
                flux = None
                for keyword in EDGE_KEYWORDS:
                    if keyword in props:
                        flux = keyword
                        break
                value = state_machine.get_value(props[flux])
                actions = props['actions'] if 'actions' in props else []
                if callable(value):
                    value = value(self)
                if value > 0 or flux == 'amount-all-but':
                    result.append((state, flux, value, cond_result, actions))
        return result

    def production_from(self, initial_state, state_machine):
        """Return a list of tuples composed of:

        - each of the possible states that can be produced from the specified
        initial state (some depending on a condition)

        - the transition rate, probability or amount to each state in
        the specified state_machine (possible clauses: 'rate',
        'proba', 'amount', 'amount-all-but' with the actual value)

        - a tuple indicating who can produce cross the edge, depending on
        conditions (either ('population', qty) or ('agents', list))

        - the prototypes of agents produced through the edge (if
        any).

        The conditions, rates and probabilities may depend on the state
        variables or properties of the current unit.

        """
        result = []
        # remove unfulfilled 'when' clauses if any
        for (state, props) in state_machine.graph.edges_from(initial_state,
                                                             type_id=EdgeTypes.PRODUCTION):
            if 'when' in props:
                when = state_machine.get_value(props['when'])
                fulfilled = when(self) if callable(when) else when
                if not fulfilled:
                    continue
            cond_result = self.get_content()
            # if any condition, evaluate it
            if 'cond' in props:
                cond = state_machine.get_value(props['cond'])
                # compute the content tuple (either ('population', qty)
                # or ('agents', lsit)) which fulfils the condition
                cond_result = self.evaluate_condition(cond)
            # only edges with condition fulfilled are taken into account
            if cond_result:
                # print(cond_result)
                flux = None
                for keyword in EDGE_KEYWORDS:
                    if keyword in props:
                        flux = keyword
                        break
                value = state_machine.get_value(props[flux])
                protos = props['prototype'] if 'prototype' in props else None
                if callable(value):
                    value = value(self)
                if value > 0 or flux == 'amount-all-but':
                    result.append((state, flux, value, cond_result, protos))
        return result


    def _compute_production(self, value, flux, pop_size, stochastic):
        """Compute the values of production edges.

        """
        # ALL PRODUCTIONS COME FROM THE SAME POPULATION
        # (reference_pop) The `values` can represent : 1) amounts - in
        # that case no transformation occurs, 2) probabilities - in
        # that case the values must be converted to rates if the
        # target compartment is deterministic, otherwise the step
        # duration must be taken into account; 3) rates - computed
        # using Poisson distribution if the target compartment is
        # stochastic
        if flux == 'amount-all-but':
            amount = max(0, pop_size - value)
        elif flux == 'amount':
            amount = value
        elif flux == 'proba':
            # if stochastic uses binomial
            if stochastic:
                # aggregate probability wrt the time step duration
                value = aggregate_probabilities([value], self.model.delta_t)[0]
                amount = np.random.binomial(pop_size, value)
            else:
                # otherwise treat proba as proportion
                amount = pop_size * value
        elif flux == 'rate':
            # if stochastic: use Poisson distribution
            if stochastic:
                amount = np.random.poisson(value * pop_size * self.model.delta_t)
                # amount = np.random.binomial(pop_size, 1-np.exp(-value * self.model.delta_t))
            else:
                # consider rate as a speed
                # S = S0 exp(rate * dt) => S += S * (exp(rate*dt) - 1)
                # amount = value * pop_size * self.model.delta_t
                amount = pop_size * (np.exp(value * self.model.delta_t) - 1)
        else:                   # should not happen !
            amount = None
        return amount


    def _compute_values_for_unique_population(self,
                                              values,
                                              flux,
                                              reference_pop,
                                              stochastic):
        """Restructure the values according to the situation, for
        edges concerning the same population.

        """
        # ALL TRANSITIONS AFFECT THE SAME POPULATION (reference_pop)
        # The `values` can represent : 1) amounts - in that case no
        # transformation occurs, 2) probabilities - in that case the
        # values must be converted to rates if the target compartment
        # is deterministic, otherwise the step duration must be taken
        # into account; 3) rates - in that case the values must be
        # converted to probabilities if the target compartment is
        # stochastic
        # print('IDENTICAL')
        available_flux = set(flux)
        # try:
        assert len(available_flux) == 1 or available_flux == set(['amount', 'amount-all-but'])
        # except:
        #     print(available_flux)

        method = None
        if 'amount' in available_flux or 'amount-all-but' in available_flux:
            # handle values as amounts
            method = 'amount'
            total_ref_pop = count_population(reference_pop)
            # check that values are between 0 and the population size,
            # if needed invert 'amount-all-but' values
            values = tuple([max(0, min(total_ref_pop-v, total_ref_pop))\
                              if f == 'amount-all-but'\
                              else max(0, min(v, total_ref_pop))
                            for (f, v) in zip(flux, values)])
            # when finished the length of values is the number of
            # outgoing edges
            # print('AMOUNT', values)
        elif 'proba' in available_flux:
            if not stochastic:
                # then if the target compartment is deterministic,
                # probabilities must be converted into rates
                # print('PROBA -> RATES', values)
                values = probabilities_to_rates(values + (1 - sum(values),))
                # when finished the length of values is the number of
                # outgoing edges
                # print(values)
            else:
                # aggregate probabilities wrt the time step duration
                values = aggregate_probabilities(values, self.model.delta_t)
                values = values + (1 - sum(values),)
                # when finished the length of values is the number of
                # outgoing edges + 1
                # print('PROBA', values)
        elif not stochastic:
            # print('RATES', values)
            pass
        else:
            # otherwise values are transformed from rates to
            # probabilities
            values = rates_to_probabilities(sum(values),
                                            values,
                                            delta_t=self.model.delta_t)
            # when finished the length of values is the number of
            # outgoing edges + 1
            # print("RATES -> PROBAS", values)
        return values, method

    def _compute_values_for_multiple_populations(self,
                                                 values,
                                                 flux,
                                                 populations,
                                                 stochastic):
        """Restructure the values according to the situation, for
        edges concerning distinct populations.

        """
        # IN THAT CASE, EACH POSSIBLE TRANSITION IS RELATED TO A
        # SPECIFIC SUBGROUP OF THE COMPART.
        ### IMPORTANT: we assume that all conditions are disjoint,
        ### i.e. there is no intersection between the populations. IF
        ### NOT THE CASE, this means that the model is not really
        ### consistent, and the calculation of probabilities should be
        ### done on each individual... thus why use a StructuredView
        ### instead of a set of EvolvingAtom ???
        # print('MULTIPLE')
        pop_sets = [SortedSet(pop[AGENTS]) for pop in populations]
        # pop_sets must be disjoint
        assert(not any([pop_sets[i] & pop_sets[j]
                        for i in range(len(pop_sets)-1)
                        for j in range(i+1, len(pop_sets))]))
        # binomial values must be computed for each transition to
        # determine how many units among the candidates will cross the
        # transition, thus we need probabilities. If all values are
        # given as rates, they must be transformed
        method = None
        available_flux = set(flux)
        # try:
        assert len(available_flux) == 1 or available_flux == set(['amount', 'amount-all-but'])
        # except:
        #     print(available_flux)

        if available_flux == {'rate'} and stochastic:
            # values are transformed from rates to probabilities
            values = rates_to_probabilities(sum(values),
                                            values,
                                            delta_t=self.model.delta_t)
            # print('RATES -> PROBAS', values)
        elif 'amount' in available_flux or 'amount-all-but' in available_flux:
            method = 'amount'
            pops = [len(p) for p in pop_sets]
            values = tuple([max(0, min(pop-v, pop))\
                              if f == 'amount-all-but'\
                              else max(0, min(v, pop))
                            for (f, v, pop) in zip(flux, values, pops)])
            values = values #+ (1 - sum(values),)
            # print('AMOUNTS:', values)
        elif 'proba' in available_flux:
            if stochastic:
                # values must be aggregated wrt the times step duration
                values = aggregate_probabilities(values,
                                                 delta_t=self.model.delta_t)
                values = values + (1 - sum(values),)
                # print('PROBAS:', values)
            else:
                # values must be transformed from probabilities to rates
                # print('PROBABILITIES -> RATES', values)
                values = probabilities_to_rates(values + (1 - sum(values),))
                # print(values)
        else:
            print('ERROR - DETERMINISTIC COMPARTMENT MODEL MUST NOT'
                  ' TRY TO HANDLE SEPARATE SUBPOPULATIONS')
        return values, method


    def evaluate_condition(self, condition):
        """Return the content (tuple either ('population', qty) or
        ('agents', list)) if the condition is fulfilled, () otherwise.

        """
        if callable(condition):
            condition = condition(self)
        return self.get_content() if condition else ()

    def evaluate_event(self, name):
        """Evaluate if the specified event name is fulfilled, using a
        calendar. The calendar is determined dynamically according to
        the name of the event.

        TAG: USER
        """
        # print(self, 'evaluating event', name, "at", self.statevars.step)
        return self.model.get_calendar_for_event(name)[name](self.statevars.step)

    def is_in_state(self, state_name):
        """Evaluate if the agent is in the specified state. The state name is
        expected to be a valid state of one of the state machines. If
        so, the name of the state machine is expected to be one of the
        statevars of the agent.

        """
        # retrieve name of the state machine associated with the state_name
        varname = self.get_model_value(state_name).state_machine.machine_name
        result = self.statevars[varname].name == state_name\
                 if varname in self.statevars else False
        # print(self, 'testing var:', varname, ' = ', result)
        return result

    # def time_before_event(self, event_name):
    #     """Return the time to wait from the current time to the specified
    #     *event_name*.
    #     """
    #     pass self.model.get_cal

    def get_outbox(self):
        """Return the content of the outbox.

        TAG: USER
        """
        return self._outbox

    def add_outbox(self, message):
        """Appendd a message to the outbox.

        TAG: USER
        """
        return self._outbox.append(message)

    def reset_outbox(self):
        """Reset the content of the outbox.

        TAG: USER
        """
        self._outbox = []

    def add_inbox(self, messages=[]):
        """Add the specified list of messages in the inbox.

        TAG: USER"""
        self._inbox += messages

    def clean_inbox(self):
        """Remove non sticky messages from the inbox.

        TAG: USER
        """
        self._inbox = [message for message in self._inbox if message.get('sticky')]

    def checkout_inbox(self):
        """Pick up agent's inbox content.

        TAG: USER
        """
        for message in self._inbox:
            for name, value in message.items():
                self.set_information(name, value)



"""
.. module:: emulsion.model.emulsion_model

.. moduleauthor:: Sébastien Picault <sebastien.picault@inra.fr>

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   pathlib                 import Path
from   datetime                import datetime, timedelta
from   copy                    import deepcopy

import yaml
import dateutil.parser         as     dup

from   sympy                   import sympify, Symbol

from   sortedcontainers        import SortedSet, SortedDict
from   jinja2                  import Environment, PackageLoader

import networkx                as     nx

from   emulsion.tools.state    import StateVarDict
from   emulsion.tools.misc     import read_from_file, load_class, create_new_serial, create_weighted_random
from   emulsion.tools.calendar import EventCalendar

from   emulsion.model.functions      import DEFAULT_LEVEL_INFO, DEFAULT_GROUPING_INFO,\
    make_function, make_when_condition
from   emulsion.model.exceptions     import SemanticException
from   emulsion.model.state_machines import StateMachine

#  ______                 _     _             __  __           _      _
# |  ____|               | |   (_)           |  \/  |         | |    | |
# | |__   _ __ ___  _   _| |___ _  ___  _ __ | \  / | ___   __| | ___| |
# |  __| | '_ ` _ \| | | | / __| |/ _ \| '_ \| |\/| |/ _ \ / _` |/ _ \ |
# | |____| | | | | | |_| | \__ \ | (_) | | | | |  | | (_) | (_| |  __/ |
# |______|_| |_| |_|\__,_|_|___/_|\___/|_| |_|_|  |_|\___/ \__,_|\___|_|


class EmulsionModel(object):
    """Class in charge of the description of a multi-level
    epidemiological model. Such models are composed of several
    processes which may take place at different
    organization/observation levels. Some of those processes are
    reified through a function implemented within an agent, while
    others are represented by a state machine. The model is also in
    charge of handling symbolic expressions and pre-computing their
    values, e.g. concerning parameters, transmission functions,
    etc.

    """
    def __init__(self, filename=None, description=None):
        """Instantiate the model either from a configuration file or
        from a string. Both must contain a YAML-based description of
        the model.

        """
        if filename:
            self.filename = filename
            self.parse(read_from_file(filename))
        else:
            self.filename = None
            self.parse(description)

    def __repr__(self):
        return '%s "%s"' % (self.__class__.__name__, self.model_name)

    def normalize_format(self):
        """Return a YAML representation of the model, reformatted to to print
        lists and dicts using the flow style (no [], no {}) instead of
        the block style, especially for syntax checking where rules
        are defined only for flow style.

        """
        return yaml.dump(self._description, default_flow_style=False)

    def copy(self):
        """Return a copy of object (naif method).
        TODO: Find a way to avoid recharging compartment class when
        intentiate a MultiProcessManager class with a old model.

        """
        return deepcopy(self)

    def _reset_all(self):
        # namespace for all symbols used in the model
        # dict string -> sympy symbol
        self._namespace = {}
        # namespace for calendar periods in the model
        # dict string -> sympy symbol
        self._event_namespace = {}
        # cache for all expressions encountered in the model
        # dict string -> sympy expression
        self._expressions = {}
        # cache for all values encountered in the model
        # dict string -> value or function
        self._values = {}
        # the original description to parse (dict from YAML document)
        self._description = {}
        # default values for modules used in symbolic computing
        self.modules = ['emulsion.tools.functions', 'numpy', 'numpy.random', 'math', 'sympy']
        # name of the model
        self.model_name = 'New Model'
        # time unit (used to specify parameter values)
        self.time_unit = 'days'
        # duration of a simulation time step (number of time_units)
        self.delta_t = 1
        # duration of a simulation time step ('true' duration)
        self.step_duration = timedelta(days=1)
        # origin date for the simulation
        self.origin_date = datetime.now()
        # dictionary of calendars (keyed by their name)
        self.calendars = SortedDict()
        # reverse dict event -> calendar name
        self._events = {}
        # dict of levels
        self.levels = {}
        # dict of aggregate vars by level
        self.aggregate_vars = {}
        # list of the processes involved in the model
        self.processes = []
        # description of the compartment associated with some of the processes
        self.compartments = {}
        # description of the state machines associated with some of the processes
        self.state_machines = {}
        # dict of all 'parameters' encountered in the model
        self.parameters = StateVarDict()
        # dict of all 'statevars' encountered in the model
        self.statevars = StateVarDict()
        # dict of all conditions encountered in the model
        self.conditions = {}
        # dict of all actions encountered in the model
        self.actions = {}
        # dict of actions to run for state initialization
        self.init_actions = {}
        # dict of all distributions encountered in the model
        self.distributions = {}
        # dict of all prototypes encountered in the model
        self.prototypes = {}
        # dict of all initial conditions encountered in the model
        self.initial_conditions = {}
        # dict of enumerate types used in special variables
        self.types = {}
        # dict of all states existing in state machines
        self.states = StateVarDict()
        # dict of parameters to log when explicitly changed
        self.params_to_log = StateVarDict()


    def add_init_action(self, machine_name, state, action):
        """Add an action to be performed when initializing agents for
        the specified state of the state machine. Mainly used for
        durations.

        """
        if machine_name not in self.init_actions:
            self.init_actions[machine_name] = {}
        if state not in self.init_actions[machine_name]:
            self.init_actions[machine_name][state] = [action]
        else:
            self.init_actions[machine_name][state].append(action)


    def _init_namespace(self):
        """Init the list of all encountered symbols, which have to be
        either parameters or state variables. The association between
        the name and the corresponding Symbol is stored in the
        namespace attribute.

        """
        # name space for 'regular' symbols (i.e. parameters, statevars)
        self._namespace.clear()

        for keyword in ['parameters', 'statevars']:
            if keyword in self._description:
                self._namespace.update({param: Symbol(param)
                                        for param in self._description[keyword]})
            else:
                self._description[keyword] = {}

    def get_value(self, name):
        """Return the value associated with the specified name. If the method
        is called with a name that is not in the model values, try to
        parse the expression, to prevent failures when called with a
        final value.

        """
        if name == "delta_t":
            return self.delta_t
        try:
            return self._values[name]
        except KeyError:
            try:
                return float(sympify(name, locals=self._namespace))
            except TypeError:
                ### cases of string parameters ?
                return name
                # result = sympify(name, locals=self._namespace)
                # print(result, type(result))
                # import sys
                # sys.exit(-1)

    def change_parameter_values(self, changes, log_params=False):
        """Naive method to change several parameter values at the same time. """
        for name, value in changes.items():
            if name == 'delta_t':
                self._description['time_info']['delta_t'] = value # was int(value) previously
            else:
                self._description['parameters'][name]['value'] = value
        self.parse(self._description)
        self.params_to_log.clear()
        if log_params:
            for name in changes:
                value = self.get_value(name) if name != 'delta_t'\
                                             else self.delta_t
                self.params_to_log[name] = value

    def set_value(self, name, value):
        """Naif method to change a parameter's value.
        Will be more efficient when treating several parameters at
        the same time.

        """
        self._description['parameters'][name]['value'] = value
        self.parse(self._description)

    def parse(self, description):
        """Build the EmulsionModel from the specified dictionary
        (expected to come from a YAML configuration file).

        """
        self._reset_all()
        # keep an exhaustive description
        self._description = description
        # retrieve the name of the model
        self.model_name = self._description['model_name']
        # build association between symbols names (all parameter and
        # statevars names) and true sympy symbols
        self._init_namespace()
        # parse time informations
        self.build_timeinfo()
        # parse output options
        self.build_outputs_options()
        #  rapidly check state machines to build is_state and
        #  duration_in_state_machine properties
        self.check_state_machines()
        # parse parameters, state variables and distributions
        self.build_parameters()
        self.build_statevars()
        self.build_levels()
        self.build_distributions()
        # parse processes
        self.build_processes()
        # parse compartment description
        self.build_compartment_desc()
        # parse state machines
        self.build_state_machines()
        # parse prototypes
        self.build_prototypes()
        # parse initial_conditions
        self.build_initial_conds()
        # parse actions
        self.build_actions()
        # calculate expressions from parameters
        self.calculate_compound_params()
        # replace expressions by values or lambdas
        self.compute_values()
        self._description['parameters']['delta_t'] = {
            'desc': 'duration of the simulation time step',
            'value': self.delta_t
        }

    def build_outputs_options(self):
        """Parse the outputs options of the model.
        The agents will treat extra variables for outputs (TODO), and
        the simulation classes will treat period outputs.

        Example of YAML specification:
        ------------------------------
        outputs:
          # level
          herd:
            period:1
          # level
          metapop:
            period: 7
            extra_vars:
              - step
              - infection_date
        """
        if 'outputs' in self._description:
            self.outputs = self._description['outputs']
        else:
            self.outputs = {}

    def build_timeinfo(self):
        """Parse the description of the model and extract the
        information related to time management, i.e. time unit,
        duration of a simulation step, origin date, calendars for
        specific events, etc.

        Example of YAML specification:
        ------------------------------
        time_info:
        time_unit: days
        delta_t: 7
        origin: 'May 1'
        calendar:
          name:
          period: {days: 52}
          events:
            spring: {begin: 'April 8', end: 'April 23'}
            summer: {begin: 'July 8', end: 'September 3'}
            national: {date: 'July 14'}
        """
        self._event_namespace.clear()
        if 'time_info' in self._description:
            tinfo = self._description['time_info']
            # compute effective duration of one time step
            self.time_unit = tinfo['time_unit']
            # parse expression for delta_t if any and store its value
            # into the delta_t attribute
            # self.delta_t = tinfo['delta_t']
            self.delta_t = float(sympify(tinfo['delta_t'],
                                         locals=self._namespace))
            timedesc = {self.time_unit: self.delta_t}
            self.step_duration = timedelta(**timedesc)
            # print(self.step_duration)
            # origin date for the simulation
            if 'origin' in tinfo:
                self.origin_date = dup.parse(tinfo['origin'])
            # total duration for the simulation (in time units)
            if 'total_duration' in tinfo:
                self.parameters['total_duration'] =\
                        sympify(tinfo['total_duration'], locals=self._namespace)
            # handle calendars
            if 'calendars' in tinfo:
                self.build_calendar(tinfo['calendars'])

    def build_calendar(self, calendar_desc):
        """Build a representation of calendars."""
        # init name space for 'event' symbols (periods of
        # time) and handle period definitions
        for cal_name, cal in calendar_desc.items():
            events = {}
            if 'events' in cal:
                cal_period = timedelta(**cal['period']) if 'period' in cal else None
                for event_name, definition in cal['events'].items():
                    # register the event name in the events namespace
                    self._event_namespace[event_name] = Symbol(event_name)
                    # if keyword "date" is used, begin = end
                    if 'date' in definition:
                        events[event_name] = (dup.parse(definition['date']),
                                              dup.parse(definition['date']))
                    else:
                        events[event_name] = (dup.parse(definition['begin']),
                                              dup.parse(definition['end']))
                        # if begin and end date are specified,
                        # register events corresponding to "begin_event"
                        # and "end_event"
                        for keyword in ['begin', 'end']:
                            event_limit = keyword + '_' + event_name
                            self._event_namespace[event_limit] = Symbol(event_limit)
                            events[event_limit] = (dup.parse(definition[keyword]),
                                                   dup.parse(definition[keyword]))
                    # register the durations of each event
                    # compute duration of event
                    begin, end = events[event_name]
                    if end >= begin:
                        dur = end - begin
                    else:
                        dur = end - begin + cal_period
                    # compute duration of 1 time unit
                    unit = {self.time_unit: 1}
                    dur_unit = timedelta(**unit)
                    if (dur_unit > timedelta(days=1)):
                        dur += timedelta(days=1)
                    else: # 1h / 1min / 1s ...
                        dur += dur_unit
                    # convert duration to time units
                    event_duration = dur / dur_unit
                    duration_param = 'duration_of_{}'.format(event_name)
                    self._namespace[duration_param] = Symbol(duration_param)
                    self._values[duration_param] = sympify(event_duration)
                    # cal_name = cal['name']
            self.calendars[cal_name] = EventCalendar(cal_name,
                                                     self.step_duration,
                                                     self.origin_date,
                                                     cal_period,
                                                     **events)
            for event in self.calendars[cal_name].get_events():
                self._events[event] = cal_name
                expression = sympify(event, locals=self._event_namespace)
                self._values[str(expression)] = make_when_condition(
                    expression, modules=self.modules)


    def get_calendar_for_event(self, name):
        """Return the calendar providing the specified event name."""
        return self.calendars[self._events[name]]

    def build_parameters(self):
        """Parse the description of the model and extract the
        parameters, either with their value, or with the expression to
        compute them.

        Example of YAML specification:
        ------------------------------
        parameters:
          p:
            desc: infection probability
            value: '(1-exp(-E_total)) * (1 - phi*vaccinated)'
          phi:
            desc: efficiency of vaccination
            value: 0.79

        """
        if 'modules' in self._description:
            self.modules += self._description['modules']
        if 'parameters' in self._description:
            for (key, val) in self._description['parameters'].items():
                self.parameters[key] = sympify(val['value'],
                                               locals=self._namespace)

    def build_statevars(self):
        """Parse the description of the model and extract the state
        variables that agents running this model must have.

        Example of YAML specification:
        ------------------------------
        statevars:
          E_total:
            desc: total bacteria deposited in the environment
          vaccinated:
            desc: 0/1 value describing the vaccination state

        """
        if 'statevars' in self._description:
            self.statevars = StateVarDict(self._description['statevars'])


    def build_distributions(self):
        """Parse the description of the model and extract the
        distributions, either with their value, or with the expression
        to compute them. A distribution is a dictionary of the form
        {quantity: probability}. It is stored as a list of tuples
        (quantity, probability) which is more convenient for
        subsequent processing.

        Example of YAML specification:
        ------------------------------
        distributions:
          - shedding_dist1:
              desc: distribution of shedding
              value:
                low_shedding: 0.85
                med_shedding: 0.15

        """
        if 'distributions' in self._description:
            for list_item in self._description['distributions']:
                for (key, val) in list_item.items():
                    self.distributions[key] = [
                        (sympify(quantity, locals=self._namespace),
                         sympify(probability, locals=self._namespace))
                        for quantity, probability in val['value'].items()]

    def build_initial_conds(self):
        """Parse the description of the model and extract the initial
        conditions for each level, either with their value, of with
        the expression to compute them. An initial condition is a
        dictionary containing a description of the 'content' of the
        level (i.e. the quantity and initial state of sub-levels):
        - either the 'population:' keyword followed by a dict State -> Amount
        - or 'prototype:' and 'amount:' keywords followed by their values

        Example of YAML specification:
        ------------------------------
        initial_conditions:
          metapop:
            - prototype: healthy # defines init_prevalence
              amount: 100
            - prototype: sick
              amount: 1
          herd:
            - population:
                infection:
                  S: 'init_pop * (1 - init_prevalence)'
                  I: 'init_pop * init_prevalence'

        """
        if 'initial_conditions' in self._description:
            for level, desc in self._description['initial_conditions'].items():
                self.initial_conditions[level] = {
                    'total': '0',
                    'populations': [],
                    'prototypes': []
                }
                for list_items in desc:
                    ## build initial conditions for compartment-based model
                    if 'population' in list_items:
                        for pop_spec in list_items['population']:
                            if 'total' in pop_spec:
                                self.initial_conditions[level]['total'] = self.add_expression(pop_spec['total'])
                            else:
                                keys = pop_spec['vars']
                                amount = pop_spec['amount']
                                if not isinstance(keys, list):
                                    keys = [keys]
                                self.initial_conditions[level]['populations'].append(
                                    (keys, self.add_expression(amount))
                                )
                    else:
                        ## build initial conditions for individual-based model
                        protos = list_items['prototype']
                        if not isinstance(protos, list):
                            protos = [protos]
                            list_items['prototype'] = protos
                        amount = list_items['amount']
                        if 'proba' in list_items:
                            probas = list_items['proba']
                            if not isinstance(probas, list):
                                probas = [probas]
                        else:
                            probas = ['1'] * len(protos)
                        list_items['proba'] = probas
                        assert(0 <= len(protos) - len(probas) <= 1)
                        self.initial_conditions[level]['prototypes'].append(
                            (protos, self.add_expression(amount), [self.add_expression(p) for p in probas]))
        # print(self._description['initial_conditions'])
        # print(self.initial_conditions)

    def build_prototypes(self):
        """Parse the description of the model and extract the list of
        prototypes for each level, either with their value, or with
        the expression to compute them. A prototype is a dictionary of
        the form {statevar: value}..

        Example of YAML specification:
        ------------------------------
        prototypes:
          animals:
            - newborn:
                desc: prototype for new calves
                health_state: M
                life_state: NP
                age: 0

        """
        if 'prototypes' in self._description:
            for level, prototypes in self._description['prototypes'].items():
                self.prototypes[level] = {}
                for list_item in prototypes:
                    for (key, val) in list_item.items():
                        self.prototypes[level][key] = {}
                        for variable, value in val.items():
                            result = value
                            if variable != 'desc':
                                if value == 'random':
                                    result = '_random_' + variable
                                elif value == 'default':
                                    result = '_default_' + variable
                                elif value == 'next_state':
                                    result = '_next_state_' + variable
                                    # associate to a function that returns the successor of the state contained in the variable (or a random state if the variable was not defined)
                                    self._values[result] = lambda agent: agent.statevars[variable].successor if variable in agent.statevars else agent.model.state_machines[variable].get_random_state()
                                elif value == 'previous_state':
                                    result = '_previous_state_' + variable
                                    # associate to a function that returns the predecessor of the state contained in the variable (or a random state if the variable was not defined)
                                    self._values[result] = lambda agent: agent.statevars[variable].predecessor if variable in agent.statevars else agent.model.state_machines[variable].get_random_state()
                                elif type(value) == str and value.startswith('random('):
                                    # result = '_weighted_random_' + variable
                                    result = '{}_{}'.format(variable, value)
                                    if value[6] == '(' and value[-1] == ')':
                                        weights = [self.add_expression(e) for e in value[7:-1].split(',')]
                                    self._values[result] = create_weighted_random(variable, weights, model=self)
                                elif type(value) == str and value.startswith('serial'):
                                    result = '_serial_' + variable
                                    if value != 'serial' and value[6] == '(' and value[-1] == ')':
                                        end = self.add_expression(value[7:-1])
                                    else:
                                        start, end = 0, None
                                    self._values[result] = create_new_serial(end=end, model=self)
                                elif value in self.statevars:
                                    self._values[result] = lambda agent: agent.statevars[result]
                                self.prototypes[level][key][variable] = result


    def get_prototype(self, level, name):
        """Return a ready-to-use prototype, i.e. a StateVarDict corresponding
        to the prototype for the specified level and name, where
        symbols associated to statevariables are already replaced by
        their value in the model. 'Hard-coded' lists are transformed
        into tuples (to be hashable).

        """
        return StateVarDict({
            var: tuple(val) if isinstance(val, list) else self.get_value(val)
            for var, val in self.prototypes[level][name].items()
        })

    def build_levels(self):
        """Parse the description of different level of simulations.
        Most of time, there are tree differents levels:
        individual, herd, metapopulation.

        Example of YAML specification:
        ------------------------------
        levels:
          individuals:
            super:
              class: AtomAgent
            class_name: Cow
          herd:
            super:
              class: MultiProcessManager
            class_name: QfeverHerd
          metapop:
            super:
              class: MultiProcessManager
              master_class: StructuredView
            class_name: QfeverMetaPop

        """
        if 'levels' in self._description:
            self.levels = self._description['levels']
            self.root_level = None
            self.levels_graph = nx.DiGraph()

        # build levels graph
        for level in self.levels:
            self.levels_graph.add_node(level)
            desc = self.levels[level]
            if 'contains' in desc:
                for sublevel in desc['contains']:
                    self.levels_graph.add_edge(level, sublevel)

        # check consistency
        if nx.is_arborescence(self.levels_graph):
            self.root_level = list(nx.topological_sort(self.levels_graph))[0]
        else:
            print('Warning, levels do not consitute an arborescence!')
            print(self.levels_graph.edges, self.levels_graph.nodes)
        # print(self.levels)

        # add default information if missing
        for level in self.levels:
            desc = self.levels[level]
            if 'module' not in desc:
                # try to build module from file
                if 'file' in desc:
                    path = Path(desc['file'])
                    desc['module'] = '.'.join(path.parent.parts + (path.stem,))
                    # if no class name build one ("DefaultLevelnameClass")
                    if 'class_name' not in desc:
                        desc['class_name'] = "Default{}Class".format(level.capitalize())
                else:
                    # if level built by aggregation use default level class
                    if 'aggregation_type' in desc:
                        desc['module'] = DEFAULT_LEVEL_INFO[desc['aggregation_type']]['level']['module']
                        desc['class_name'] = DEFAULT_LEVEL_INFO[desc['aggregation_type']]['level']['class_name']
                    else:
                        # try to retrieve upper level if any
                        predecessors = list(self.levels_graph.predecessors(level))
                        if not predecessors:
                            desc['module'] = DEFAULT_LEVEL_INFO['IBM']['sublevels']['module']
                            desc['class_name'] = DEFAULT_LEVEL_INFO['IBM']['sublevels']['class_name']
                        else: # take first information regarding aggregation type
                            ## TODO (research topic on multi-level
                            ## design patterns): hierarchize possibly
                            ## different aggregation types from upper
                            ## levels
                            for pred in predecessors:
                                desc_pred = self.levels[pred]
                                if not 'aggregation_type' in desc_pred:
                                    continue
                                pred_info = DEFAULT_LEVEL_INFO[desc_pred['aggregation_type']]
                                if 'sublevels' not in pred_info:
                                    continue
                                desc['module'] = pred_info['sublevels']['module']
                                desc['class_name'] = pred_info['sublevels']['class_name']
                            if 'module' not in desc:
                                desc['module'] = None
                                desc['class_name'] = None
            if 'aggregation_type' in desc:
                default = DEFAULT_LEVEL_INFO[desc['aggregation_type']]
                if 'super' not in desc:
                    desc['super'] = default['level']
                if not desc['module'].startswith('emulsion.agent'):
                    if 'sublevels' in default:
                        if 'contains' in desc:
                            for sublevel in desc['contains']:
                                if ('module' not in self.levels[sublevel] or\
                                    (not self.levels[sublevel]['module'].startswith('emulsion.agent')))\
                                   and ('super' not in self.levels[sublevel]):
                                    self.levels[sublevel]['super'] = default['sublevels']
                if 'aggregate_vars' in desc:
                    if level not in self.aggregate_vars:
                        self.aggregate_vars[level] = []
                    for agg_desc in desc['aggregate_vars']:
                        newvar = agg_desc['name']
                        sourcevar = agg_desc['collect']
                        oper = agg_desc['operator'] if 'operator' in agg_desc else 'sum'
                        self.aggregate_vars[level].append((newvar, sourcevar, oper))
            # print(self.aggregate_vars)

    def get_agent_class_for_level(self, level):
        return load_class(module=self.levels[level]['module'],
                          class_name=self.levels[level]['class_name'])

    def build_processes(self):
        """Parse the description of the model and retrieve the list of
        processes with different level.

        Example of YAML specification:
        ------------------------------
        processes:
          herd:
            - bacterial_dispersion
            - culling_process
            - infection
          metapop:
            - inbox_distribution
            - outbox_distribution
        """
        if 'processes' in self._description:
            self.processes = self._description['processes']

    def check_state_machines(self):
        """Rapidly inspect the name of state machines and of their states to
        pre-register automatically built statevars.

        """
        if 'state_machines' in self._description:
            for machine_name, description in self._description['state_machines'].items():
                self.state_machines[machine_name] = None # to build later
                desc = "duration elapsed (time units) in current {} state".format(machine_name)
                prop_name = 'duration_in_{}'.format(machine_name)
                # self._description['statevars'][prop_name] = {'desc': desc}
                self.statevars[prop_name] = {'desc': desc}
                self._description['statevars'][prop_name] = {'desc': desc}
                self._namespace[prop_name] = Symbol(prop_name)
                desc = "time step when agent entered current {} state".format(machine_name)
                prop_name = '_time_entered_{}'.format(machine_name)
                self.statevars[prop_name] = {'desc': desc}
                self._description['statevars'][prop_name] = {'desc': desc}
                self._namespace[prop_name] = Symbol(prop_name)
                desc = "time step after which agent is allowed to leave current {} state".format(machine_name)
                prop_name = '_time_to_exit_{}'.format(machine_name)
                self.statevars[prop_name] = {'desc': desc}
                self._description['statevars'][prop_name] = {'desc': desc}
                self._namespace[prop_name] = Symbol(prop_name)
                for statedict in description['states']:
                    for state_name in statedict:
                        desc = "Test if {} = {}".format(machine_name, state_name)
                        prop_name = 'is_{}'.format(state_name)
                        self._description['statevars'][prop_name] = {'desc': desc}
                        # self.statevars[prop_name] = desc
                        self._namespace[prop_name] = Symbol(prop_name)
            # print(self._namespace, self.statevars)


    def build_state_machines(self):
        """Parse the description of the model and build all the
        required state machines.

        """
        if 'state_machines' in self._description:
            for machine_name, description in self._description['state_machines'].items():
                self.state_machines[machine_name] = StateMachine(machine_name,
                                                                 description,
                                                                 self)

    def build_compartment_desc(self):
        """Inspect the `grouping` part of the model (if
        any) in order to store the corresponding information.

        """
        if 'grouping' in self._description:
            self.compartments = self._description['grouping']
            for level in self.levels.keys():
                if level not in self.compartments:
                    self.compartments[level] = {}
                if 'aggregation_type' in self.levels[level]:
                    agg_type = self.levels[level]['aggregation_type']
                    if agg_type in DEFAULT_GROUPING_INFO:
                        default = DEFAULT_GROUPING_INFO[agg_type]
                        for desc in self.compartments[level].values():
                            if 'compart_manager' not in desc:
                                desc['compart_manager'] = default['compart_manager']\
                                  if  'machine_name' in desc\
                                  else default['fallback_view']
                            if 'compart_class' not in desc:
                                desc['compart_class'] = default['compart_class']
            # print(self.compartments)

    def calculate_compound_params(self):
        """Inspect all edges of the health states graph and compute
        the actual probabilities associated to expressions.
        ### TODO: check that the ultimate symbols are declared properties.

        """
        for cond, expr in self.conditions.items():
            self.conditions[cond] = self.expand_expression(expr)


    def add_expression(self, expression):
        """Add the specified expression to the dictionary of known
        expressions.

        """
        if expression not in self.parameters\
           and expression not in self.statevars\
           and expression not in self._expressions:
            self._expressions[expression] = self.expand_expression(expression)
        return expression

    def expand_expression(self, expression):
        """Transform the specified expression so as to replace all
        parameters by actual values or state variables or
        attributes.

        """
        ### WARNING: expand_expressions should iterate over all
        ### expressions at the same time (halting when no change
        ### occurs) instead of iterating over each expression one by
        ### one
        result = sympify(expression, locals=self._namespace)
        expr = result
        symbs = {s: self.parameters[s.name]
                 for s in expr.free_symbols
                 if s.name in self.parameters}
        while symbs:
            result = expr.subs(symbs)
            expr = sympify(result, locals=self._namespace)
            symbs = {s: self.parameters[s.name]
                     for s in expr.free_symbols
                     if s.name in self.parameters}
        return result

    def build_actions(self):
        """Parse the description of the model and extract the actions
        that agents must have.

        Example of YAML specification:
        ------------------------------
        actions:
          say_hello:
            desc: action performed when entering the S state

        """
        if 'actions' in self._description:
            self.actions = StateVarDict(self._description['actions'])

    def compute_values(self):
        """Check parameters and calculated compound parameters, so as
        to make them computable. In the case of parameters, number
        values are left unchanged, but strings (representing an
        expression) are replaced by a function. Regarding calculated
        compound parameters, expressions corresponding to numbers are
        converted to float values, while other expressions are
        replaced by a function.

        """
        # collect 'true' values in the parameters
        for (param, expression) in self.parameters.items():
            value = self.expand_expression(expression)
            try:
                self._values[param] = float(value)
            except:
                self._values[param] = make_function(value,
                                                    modules=self.modules)
        for (param, expression) in self._expressions.items():
            value = self.expand_expression(expression)
            try:
                self._values[param] = float(value)
            except:
                self._values[param] = make_function(value,
                                                    modules=self.modules)
        for (cond, expression) in self.conditions.items():
            value = self.expand_expression(expression)
            if any([str(symb) in self.statevars
                    for symb in expression.free_symbols]):
                self._values[cond] = make_function(value,
                                                   dtype=bool,
                                                   modules=self.modules)
            else:
                self._values[cond] = bool(value)
        for (statename, state) in self.states.items():
            self._values[statename] = state

    def get_modifiable_parameters(self):
        """Return a dictionary containing all true parameters with their
        value.

        """
        true_params =  {p: self.get_value(p)
                        for p in self.parameters
                        if not callable(self.get_value(p))}
        true_params['delta_t'] = self.delta_t
        return true_params

    def describe_parameter(self, name):
        """Return the description of the parameter with the specified
        name.

        """
        # pretty_name = pretty(sympify(name, locals=self._namespace))
        # param_name = name if pretty_name == name\
        #              else "{} ({})".format(pretty_name, name)
        param_name = name
        return "{} [parameter]:\n\t{: <72}\n\t{}".format(
            param_name,
            self._description['parameters'][name]['desc'],
            self._description['parameters'][name]['value']
        )

    def describe_variable(self, name):
        """Return the description of the statevar with the specified
        name.

        """
        # pretty_name = pretty(sympify(name, locals=self._namespace))
        # var_name = name if pretty_name == name\
        #            else "{} ({})".format(pretty_name, name)
        var_name = name
        return "{} [variable]:\n\t{: <72}".format(
            var_name,
            self._description['statevars'][name]['desc'],
        )

    def describe_name(self, name):
        """Return the description of the specified name in the model
        name.

        """
        if name in self._description['parameters']:
            return self.describe_parameter(name)
        elif name in self._description['statevars']:
            return self.describe_variable(name)
        else:
            return "{}: UNKNOWN NAME".format(name)

    def write_dot(self, parent_dir):
        """Write the graph of the each state machine in the
        specified directer name, according to the dot/graphviz format.

        """
        for name, statemachine in self.state_machines.items():
            name = self.model_name + '_' + name + '.dot'
            path = str(Path(parent_dir, name))
            statemachine.write_dot(path)

    def generate_skeleton(self, module):
        """Output a code skeleton to help writing specific pieces of code for
        the specified module to make the model work under Emulsion.

        """
        env = Environment(
            loader=PackageLoader('emulsion', 'templates'),
            extensions=['jinja2.ext.do']
        )
        template = env.get_template('specific_code.py')
        output = template.render(model=self, src_module=module)
        return output

EmulsionModel.get_modifiable_parameters.__USER_METHOD__ = ['Introspection']
EmulsionModel.describe_parameter.__USER_METHOD__ = ['Introspection']



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Tools for parallel computing.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import multiprocessing           as     mp

from   emulsion.tools.simulation import MultiSimulation, SensitivitySimulation

def job_dist(total_task, workers):
    """ Return a distribution of each worker need to do.

    """
    nb = total_task // workers
    rest = total_task % workers
    return [nb+1 if i <= rest-1 else nb for i in range(workers)]

def job(target_simulation_class, proc, **others):
    """ Simple job for a simple processes

    """
    simu = target_simulation_class(proc=proc, **others)
    simu.run()
    simu.write_dot()
    # simu.counts_to_csv()

def parallel_multi(target_simulation_class=MultiSimulation, nb_simu=None,
                   nb_proc=1, **others):
    """ Parallel loop for distributing tasks in different processes

    """
    list_proc = []
    list_nb_simu = job_dist(nb_simu, nb_proc)

    for i in range(nb_proc):
        others['nb_simu'] = list_nb_simu[i]
        others['start_id'] = sum(list_nb_simu[:i])
        proc = mp.Process(target=job, args=(target_simulation_class, i),
                          kwargs=others)
        list_proc.append(proc)
        proc.start()

    for proc in list_proc:
        proc.join()

def parallel_sensi(target_simulation_class=SensitivitySimulation, nb_proc=1,
                   **others):
    """ Parallel loop for distributing sensitivity tasks in different processes

    """
    list_proc = []
    list_nb_simu = job_dist(len(others['df']), nb_proc)

    for i in range(nb_proc):
        others['nb_multi'] = list_nb_simu[i]
        others['start_id'] = sum(list_nb_simu[:i])
        proc = mp.Process(target=job, args=(target_simulation_class, i),
                          kwargs=others)
        list_proc.append(proc)
        proc.start()

    for proc in list_proc:
        proc.join()



""".. module:: emulsion.tools.misc

Collection of various useful functions used in EMULSION framework,
especially regarding introspection.

"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   functools                 import partial
from   importlib                 import import_module
from   typing                    import List, Iterable, Tuple, Union, Any

import yaml

import numpy                     as np

from   sortedcontainers          import SortedSet


AGENTS = 1
"""Constant value for the index where list of agents are stored in
'populations' tuples.

"""

POPULATION = 1
"""Constant value for the index where population amounts are stored in
'populations' tuples.

"""

## TODO split module according to functions usage

def load_class(module=None, class_name: str = None, options: dict = {}):
    """Dynamically load the class with the specified *class_name* from the
    given *module*.

    Parameters
    ----------
    module:
        a Python module, where the class is expected to be located.
    class_name: str
        the name of the class to load
    options:
        some options

    Returns
    -------
    tuple:
        a tuple composed of the class (type) and the options.

    Todo
    ----
    - clarify the role of *options*
    """
    mod = import_module(module)
    return getattr(mod, class_name), options

def load_module(module_name: str):
    """Dynamically load the module with the specified *module_name* and
    return it.

    Parameters
    ----------
    module_name: str
        the name of a valid Python module (accessible in the
        ``PYTHONPATH`` environment variable).

    Returns
    -------
    ref:
        A reference to the Python module.

    """
    return import_module(module_name)

def rates_to_probabilities(total_rate: float, rate_values: List[float],
                           delta_t: float = 1) -> List[float]:
    """Transform the specified list of *rate_values*, interpreted as
    outgoing rates, into probabilities, according to the specified
    time step (*delta_t*) and normalized by *total_rate*.

    For exit rates :math:`\\rho_i` (one of the *rate_values*), the
    probability to stay in current state is given by:

    .. math::
        p_0 = e^{-\\delta t.\\sum_i \\rho_i}

    Thus, each rate :math:`\\rho_i` corresponds to a probability

    .. math::
        p_i = \\frac{\\rho_i}{\\sum_i \\rho_i} (1 - p_0)

    Parameters
    ----------
    total_rate: float
        the total exit rate, used for normalization purposes. If
        *rate_values* represent all possible exit rates, *total_rate*
        is their sum.
    rate_values: list
        the list of rates to transform
    delta_t: float
        the value of the time step (expressed in time units)

    Returns
    -------
    list:
        the list of probabilities corresponding to the *rate_values*.

    See Also
    --------
        `emulsion.tools.misc.probabilities_to_rates`_

    """
    # compute the exit probability
    base_proba = 1 - np.exp(- total_rate * delta_t)
    # normalize values proportionnally to the rate
    proba_values = [base_proba * rate / total_rate
                    for rate in rate_values]
    # add the probability to stay in the current state
    proba_values.append(1 - base_proba)
    return proba_values

# This function is useful to modellers for the following purposes:
rates_to_probabilities.__USER_FUNCTION__ = ['Rates / probabilities']


def aggregate_probability(probability: float, delta_t: float) -> float:
    """Transform the specified *probability* value, intended to represent
    a probability for events tested each time unit, into the
    probability for the specified time step (*delta_t*).

    Parameters
    ----------
    probability: float
        the probability value of an event (during 1 time unit)
    delta_t: float
        the value of the time step (expressed in time units)

    Returns
    -------
    float:
        the probability of the event during *delta_t* time units.
    """
    return 1 - (1 - probability)**delta_t

# This function is useful to modellers for the following purposes:
aggregate_probability.__USER_FUNCTION__ = ['Rates / probabilities']


def aggregate_probabilities(probability_values: Iterable[float],
                            delta_t: float) -> Iterable[float]:
    """From the specified *probability_values*, intended to represent
    probabilities of events duting one time unit, compute the
    probabilities for the specified time step (*delta_t*).

    Parameters
    ----------
    probability_values: list
        a list of probability values for several events (during 1 time unit)
    delta_t: float
        the value of the time step (expressed in time units)

    Returns
    -------
    list:
        the probabilities of the same events during *delta_t* time units.
    """
    return probability_values if delta_t == 1\
        else tuple(aggregate_probability(p, delta_t)
                   for p in probability_values)

# This function is useful to modellers for the following purposes:
aggregate_probabilities.__USER_FUNCTION__ = ['Rates / probabilities']


def probabilities_to_rates(probability_values: Iterable[float]) -> List[float]:
    """Transform a list of probabilities into a list of rates. The
    last value is expected to represent the probability of staying in
    the current state.

    Parameters
    ----------
    probability_values: list
        a list of probabilities, the last one representing the
        probability to stay in the current state

    Returns
    -------
    list:
        a list of rates corresponding to those probabilities.

    See Also
    --------
        `emulsion.tools.misc.rates_to_probabilities`_

    """
    if probability_values[-1] == 1:
        return [0] * (len(probability_values) - 1)
    sum_of_rates = - np.log(probability_values[-1])
    proba_change = 1 - probability_values[-1]
    values = [v * sum_of_rates / proba_change for v in probability_values]
    return values[:-1]

# This function is useful to modellers for the following purposes:
probabilities_to_rates.__USER_FUNCTION__ = ['Rates / probabilities']


# internal usage (TODO: make private to relevant module)
def rewrite_keys(name, position, change_list):
    prefix = name[:position]
    suffix = name[position+1:]
    return [(prefix + (key,) + suffix, value)
            for key, value in change_list]


def count_population(agents_or_pop: Tuple) -> Union[int, float]:
    """Return the amount of atoms represented in *agents_or_pop*.

    Parameters
    ----------
    agents_or_pop: tuple
      either ('population', qty) or ('agents', list of agents)

    Returns
    -------
    int or float:
        the amount corresponding to the population: generally, an int
        value, but deterministic models produce float values.

    """
    return agents_or_pop[POPULATION]\
      if 'population' in agents_or_pop\
      else len(agents_or_pop[AGENTS])


def select_random(origin: Iterable, quantity: int,
                  exclude: SortedSet = SortedSet()) -> List:
    """Return a random selection of *quantity* agents from the *origin*
    group, avoiding those explicitly in the *exclude* set. If the
    *origin* population proves too small, all available agents are
    taken, irrespective to the *quantity*.

    Parameters
    ----------
    origin: iterable
        the population where agents must be selected
    quantity: int
        the number of agents to select in the population
    exclude: set
        agents which are not available for the selection

    Returns
    -------
    list:
        a list of randomly selected agents according to the above constraints.
    """
    content = [unit for unit in origin if unit not in exclude]
    size = len(content)
    np.random.shuffle(content)
    return content[:min(quantity, size)]

# This function is useful to modellers for the following purposes:
select_random.__USER_FUNCTION__ = ['Selecting agents']


def read_from_file(filename: str):
    """Read the specified YAML *filename* and return the corresponding
    Python document.

    Parameters
    ----------
    filename: str
        the name of the YAML file to load

    Returns
    -------
    object:
        a Python object built by parsing of the YAML file, i.e. either
        a dict, list, or even str/int/etc... (Most YAML document will
        produce dictionaries.)

    """
    with open(filename, 'r') as fil:
        description = yaml.safe_load(fil)
    return description


def retrieve_value(value_or_function, agent):
    """Return a value either directly given by parameter
    *value_or_function* if it is a true value, or computed from this
    parameter seen as a function, with the specified *agent* as argument.

    Parameters
    ----------
    value_or_function:
        either a callable (function that applies to an agent to
        retrieve an individual value), or the value itself
    agent:
        the agent to use as parameter of the callable if necessary

    Returns
    -------
    value:
        the expected value (agent-based or not).

    """

    return value_or_function(agent)\
        if callable(value_or_function)\
        else value_or_function


def moving_average(values, window_size, mode='same'):
    """Compute a moving average of the specified *values* with respect to
    the *window_size* on which the average is calculated. The return
    moving average has the same size as the original values. To avoid
    boundary effects, use ``mode='valid'``, which produce a result of
    size ``len(values) - window_size + 1``.

    Parameters
    ----------
    values: array-like
        contains the values for moving average computation
    window_size: int
        width of the moving average window
    mode: str
        a parameter for `numpy.convolve`

    Returns
    -------
    nd_array:
        a numpy ``nd_array`` containing the values of the moving average.

    """
    window = np.ones(int(window_size)) / float(window_size)
    return np.convolve(values, window, mode)

# This function is useful to modellers for the following purposes:
moving_average.__USER_FUNCTION__ = ['Computations']


def add_new_property(agent, property_name, getter_function):
    """Add a new property to an agent.

    Actually, the property is added to the class of the agent, as
    Python properties are descriptors. Yet, the dynamic attribution of
    properties must me done through instances rather than classes,
    since agents must add the name of the property to their
    ``_mbr_cache`` attribute.

    Parameters
    ----------
    agent: AbstractAgent
        the agent to which the property must be added
    property_name: str
        the name of the property
    getter_function: lambda
        the function upon which the property is built

    """
    setattr(agent.__class__, property_name, property(getter_function))
    agent._mbr_cache.add(property_name)
    # print('Agent {} adding property {} to class {}'.format(agent, property_name, agent.__class__))


def create_population_getter(process_name: str,
                             group_name: Union[str, Tuple]):
    """Build a getter for property of the form ``total_X_Y`` where ``X``
    and ``Y`` are states of two different states machines used in a
    grouping.

    The functions can handle groupings with an arbitrary number of states.

    Parameters
    ----------
    process_name : str
        the name of the process associated with the grouping for which
        the population getter is created
    group_name : str or tuple
        either a string representing the state name (if only one), or
        a tuple of several state names

    Returns
    -------
    callable :
        A function which can be applied to an `AbstractProcessManager`
        agent which returns the population size for the specified
        group.

    """
    if not isinstance(group_name, tuple):
        group_name = (group_name,)
    # print('adding automatic property for', process_name, group_name)
    return lambda self: self.get_group_population(process_name, group_name)

def create_counter_getter(machine_name, state_name):
    return lambda self: self.counters[machine_name][state_name]


def create_state_tester(state_name):
    return lambda self: self.is_in_state(state_name)

def create_duration_getter(machine_name):
    # print('adding automatic property for', process_name, group_name)
    return lambda self: self.duration_in_current_state(machine_name)

def add_all_test_properties(agent):
    for machine in agent.model.state_machines.values():
        add_new_property(agent, 'duration_in_{}'.format(machine.machine_name),
                         create_duration_getter(machine.machine_name))
        for state in machine.states:
            add_new_property(agent, 'is_{}'.format(state.name),
                             create_state_tester(state.name))


def find_operator(operator: str):
    """Return an aggregation function named *operator*.

    Search starts with emulsion functions (module
    `emulsion.tools.functions`), which includes Python built-ins, then
    in `numpy`.

    A special shortcut is provided for percentiles: ``percentileXX``
    is interpreted as the partial function
    ``numpy.percentile(q=int(XX))``.

    Parameters
    ----------
    operator: str
        the name of the aggregation operator

    Returns
    -------
    lambda:
        a function that takes a list (or array-like) as input and
        returns the application of the *operator* to the values.

    """
    op = None
    current_module = import_module('emulsion.tools.functions')
    # search in emulsion functions module first
    if hasattr(current_module, operator):
        op = getattr(current_module, operator)
    elif operator.startswith('percentile'):
        # special case of 'percentileXX' where XX is a number (0 to
        # 100) and automatically extracted
        op = partial(np.percentile, q=int(operator[10:]))
    else:
        # otherwise search in numpy
        op = getattr(np, operator)
    return op

def create_aggregator(sourcevar: str, operator: str):
    """Create an aggregator function to be used as property getter. This
    aggregator function has to collect all values of *sourcevar* for
    agents contained in a given host, and reduce them to one avalue
    using the specified *operator*.

    Parameters
    ----------
    sourcevar: str
        the name of the variable to collect in the sublevel
    operator: str
        the name of the operator to apply to the collected values

    Returns
    -------
    lambda:
        A function which can be applied to a `MultiProcessManager`
        agent (i.e. with explicit sublevel agents) which returns the
        aggregated values for the whole population.

    """
    op = find_operator(operator)
    return lambda self: op([sublevel.get_information(sourcevar)
                            for sublevel in self['MASTER']])

def create_group_aggregator(sourcevar: str, operator: str, process_name: str,
                            group_name: Union[str, Tuple]):
    """Build a getter for property of the form ``newvar_X_Y`` where ``X``
    and ``Y`` are states of two different states machines used in a
    grouping, ``newvar`` is an aggregate variable based on collecting
    all values of *sourcevar* for the specific *group_name* and
    aggregating the values using *operator*. The functions can handle
    groupings with an arbitrary number of states.

    Parameters
    ----------
    sourcevar: str
        the name of the variable to collect in the sublevel
    operator: str
        the name of the operator to apply to the collected values
    process_name: str
        the name of the process associated with the grouping for which
        the population getter is created
    group_name: str or tuple
        either a string representing the state name (if only one), or
        a tuple of several state names

    Returns
    -------
    lambda:
        A function which can be applied to an `MultiProcessManager`
        agent (i.e. with explicit sublevel agents) which returns the
        aggregated value for the specified group.

    """
    op = find_operator(operator)
    if not isinstance(group_name, tuple):
        group_name = (group_name,)
    # print('adding automatic property for', process_name, group_name)
    def aggregate_group(self):
        try:
            return op([sublevel.get_information(sourcevar)
                       for sublevel in self.get_group_atoms(process_name, group_name)])
        except ValueError:
            return np.nan
    return aggregate_group

def create_atoms_aggregator(sourcevar: str, operator: str,
                            machine_name: str, state_name: str):
    """Build a getter for property of the form ``newvar_X`` where ``X`` if
    the value of *state_name* (a state of *state_machine*), ``newvar``
    is an aggregate variable based on collecting all values of
    *sourcevar* for the specific *group_name* and aggregating the
    values using *operator*. This function is intended to work on
    `IBMProcessManager` agents, which do not benefit from groupings.

    Parameters
    ----------
    sourcevar: str
        the name of the variable to collect in the sublevel
    operator: str
        the name of the operator to apply to the collected values
    machine_name: str
        the name of the state machine for which the getter is created
    state_name: str
        the state name

    Returns
    -------
    lambda:
        A function which can be applied to an `IBMProcessManager`
        agent (i.e. with explicit but ungrouped sublevel agents) which
        returns the aggregated value for the specified group.

    """
    op = find_operator(operator)
    def aggregate_atoms(self):
        try:
            return op([sublevel.get_information(sourcevar)
                            for sublevel in self.select_atoms(machine_name, state=state_name)])
        except ValueError:
            return np.nan
    return aggregate_atoms

def serial(start=0, end=None, model=None):
    """A very simple serial number generator."""
    value = start
    while True:
        yield value
        value += 1
        if end is not None:
            if value >= model.get_value(end):
                value = start


def create_new_serial(end=None, model=None):
    """Create the serial number generator associated to the specified
    variable.

    """
    generator = serial(start=0, end=end, model=model)
    return lambda self: next(generator)


def create_weighted_random(machine_name, weights, model=None):
    """Create a random choice function which returns a random state from
    the given state machine (among non-autoremove states), according
    to the weights. Weights are interpreted either directly as
    probabilities (if the number of weights is stricly one below the
    number of available states, the last state getting the complement
    to 1), or as true weights which are then normalized to be used as
    probabilities.

    Parameters:
    -----------
    machine_name: str
        the name of the state machine where the states must be chosen
        among the *N* non-autoremove states.
    weights: list
        a list of *N* or *N-1* model expressions assumed to produce positive numbers
    model: EmulsionModel
        the model where this function is defined

    Returns
    -------
    lambda: a function that returns a random state according to the
        values of the weights list, interpreted either as
        probabilities (if size *N-1*) or as weights (if size *N*)
        which are then normalized to provide probabilities

    """
    machine = model.state_machines[machine_name]
    states = [state for state in machine.states if not state.autoremove]
    assert(len(states) - len(weights) <= 1)
    def weighted_random(agent):
        values = [agent.get_model_value(weight) for weight in weights]
        total = sum(values)
        if len(values) < len(states):
            assert(0 <= total <= 1)
            probas = values + [1 - total]
        else:
            probas = [v / total for v in values]
        return np.random.choice(states, p=probas)
    return weighted_random



"""A Python implementation of the EMuLSion framework (Epidemiologic
MUlti-Level SImulatiONs).

Classes and functions for entities management.
"""


# EMULSION (Epidemiological Multi-Level Simulation framework)
# ===========================================================
# 
# Contributors and contact:
# -------------------------
# 
#     - Sébastien Picault (sebastien.picault@inrae.fr)
#     - Yu-Lin Huang
#     - Vianney Sicard
#     - Sandie Arnoux
#     - Gaël Beaunée
#     - Pauline Ezanno (pauline.ezanno@inrae.fr)
# 
#     INRAE, Oniris, BIOEPAR, 44300, Nantes, France
# 
# 
# How to cite:
# ------------
# 
#     S. Picault, Y.-L. Huang, V. Sicard, S. Arnoux, G. Beaunée,
#     P. Ezanno (2019). "EMULSION: Transparent and flexible multiscale
#     stochastic models in human, animal and plant epidemiology", PLoS
#     Computational Biology 15(9): e1007342. DOI:
#     10.1371/journal.pcbi.1007342
# 
# 
# License:
# --------
# 
#    Copyright 2016 INRAE and Univ. Lille
# 
#    Inter Deposit Digital Number: IDDN.FR.001.280043.000.R.P.2018.000.10000
# 
#    Agence pour la Protection des Programmes,
#    54 rue de Paradis, 75010 Paris, France
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from   collections               import OrderedDict

from   emulsion.agent.core       import EmulsionAgent
from   emulsion.agent.process    import StateMachineProcess, MethodProcess

#          _                                           _
#     /\  | |                    /\                   | |
#    /  \ | |_ ___  _ __ ___    /  \   __ _  ___ _ __ | |_
#   / /\ \| __/ _ \| '_ ` _ \  / /\ \ / _` |/ _ \ '_ \| __|
#  / ____ \ || (_) | | | | | |/ ____ \ (_| |  __/ | | | |_
# /_/    \_\__\___/|_| |_| |_/_/    \_\__, |\___|_| |_|\__|
#                                      __/ |
#                                     |___/

class AtomAgent(EmulsionAgent):
    """The AtomAgent is aimed at representing an 'individual', i.e. the
    smallest organization level to be modeled as an entity in the
    simulation. An AtomAgent may be situated in several hosts, each one
    associated with a specific tuple of state variables.

    """
    def __init__(self, **others):
        super().__init__(**others)
        self.statevars.population = 1
        self.stochastic = True
        self._host = OrderedDict()
        if 'host' in others:
            self.add_host(others['host'])

    def __len__(self):
        return 1

    def get_content(self):
        """Return the population (1) of the current unit.

        """
        return ('population', 1)

    def add_host(self, host):
        """Add the specified host to the current AtomAgent, associated
        with the specified key.

        """
        self._host[host.keys] = host
        self.simulation = host.simulation

    def remove_host(self, host, keys=None):
        """Remove the specified host from the current AtomAgent,
        associated with the specified key.

        """
        if keys is None:
            del self._host[host.keys]
        else:
            if keys in self._host:
                del self._host[keys]

    def get_host(self, key='MASTER'):
        """Retrieve the host of the current AtomAgent identified by the
        specific key.

        """
        return self._host[key]

    def clone(self, prototype=None, custom_prototype=None, **others):
        """Make a copy of the current compartment with the specified
        observable/value settings. If a prototype is provided, it is
        applied to the new atom.

        """
        new_atom = self.__class__.from_dict(self.statevars)
        new_atom.model = self.model
        new_atom.level = self.level
        new_atom.statevars.update(**others)
        if prototype is not None:
            new_atom.apply_prototype(name=prototype)
        elif custom_prototype is not None:
            new_atom.apply_prototype(prototype=custom_prototype)
        return new_atom


#  ______          _       _                     _
# |  ____|        | |     (_)               /\  | |
# | |____   _____ | |_   ___ _ __   __ _   /  \ | |_ ___  _ __ ___
# |  __\ \ / / _ \| \ \ / / | '_ \ / _` | / /\ \| __/ _ \| '_ ` _ \
# | |___\ V / (_) | |\ V /| | | | | (_| |/ ____ \ || (_) | | | | | |
# |______\_/ \___/|_| \_/ |_|_| |_|\__, /_/    \_\__\___/|_| |_| |_|
#                                   __/ |
#                                  |___/

class EvolvingAtom(AtomAgent):
    """An EvolvingAtom is able to change state according to its
    own statemachines.

    """
    def __init__(self, statemachines=[], **others):
        super().__init__(**others)
        self.processes = []
        self.statemachine_processes = {}
        self.method_processes = {}
        if statemachines:
            self.set_statemachines(statemachines)

    def set_statemachines(self, statemachines):
        """Define the state machines that this agent is able to execute."""
        self.statemachine_processes = {
            sm.machine_name: StateMachineProcess(sm.machine_name, self, sm)
            for sm in statemachines
        }

    def init_level_processes(self):
        """Initialize the level of the agent."""
        if self.level in self.model.processes:
            self.processes = self.model.processes[self.level]
            for process in self.processes:
                if process not in self.statemachine_processes:
                    self.add_method_process(process)

    def add_method_process(self, process_name, method=None):
        """Add a process based on a method name."""
        # print('process:', process_name)
        if method is None:
            method = getattr(self, process_name)
        self.method_processes[process_name] = MethodProcess(process_name, method)

    def get_machine(self, name):
        """Return the state machine with the specified name."""
        return self.statemachine_processes[name].state_machine

    def evolve(self, machine=None):
        super().evolve(machine=machine)
        self.evolve_states()

    def evolve_states(self, machine=None):
        """Change the state of the current unit according to the
        specified state machine name. If no special state machine is
        provided, executes all the machines.

        """
        # retrieve the iterable containing the processes to execute
        if machine is not None:
            self.statemachine_processes[machine].evolve()
        else:
            for process_name in self.processes:
                process = self.statemachine_processes[process_name]\
                            if process_name in self.statemachine_processes\
                            else self.method_processes[process_name]
                process.evolve()



""".. module:: emulsion.tools.simulation

Tools for providing generic simulation classes.

"""

#[HEADER]


import abc
from   pathlib        import Path
from   os.path        import exists
from   typing         import List, Iterable, Tuple, Union

import numpy          as     np
import pandas         as     pd

from   tqdm           import trange, tqdm   # a nice replacement for progressbar2

from   sqlalchemy     import create_engine

from   emulsion.model import EmulsionModel

#   ____        _               _   __  __
#  / __ \      | |             | | |  \/  |
# | |  | |_   _| |_ _ __  _   _| |_| \  / | __ _ _ __   __ _  __ _  ___ _ __
# | |  | | | | | __| '_ \| | | | __| |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |__| | |_| | |_| |_) | |_| | |_| |  | | (_| | | | | (_| | (_| |  __/ |
#  \____/ \__,_|\__| .__/ \__,_|\__|_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
#                  | |                                        __/ |
#                  |_|                                       |___/

class OutputManager(object):
    """Manager to handle different outputs (csv, database,... etc)

    """
    def __init__(self, model=None, output_dir='', output_file='counts.csv', log_file='log.txt'):
        """Initialize the output manager, specifying the model and the
        directory where outputs will be stored.

        """
        self.model = model
        self.output_dir = output_dir
        self.output_file = output_file
        self.log_file = log_file

        # database engine
        self.engine = None
        self.counts_path = None
        self.log_path = str(Path(self.output_dir, self.log_file))
        # We choose csv file by default
        self.output_type = 'csv'
        self.update_output_type()
        self.update_output_information()

    def update_output_type(self):
        """Update output type if specified in model

        """
        if 'type' in self.model.outputs:
            self.output_type = self.model.outputs['type']

    def update_output_information(self):
        """Update csv file path or database connection engine

        """
        if self.output_type == 'database':
            database_information = self.model.outputs['database_information']

            host = database_information['server_name']
            port = database_information['port']
            host = '{}:{}'.format(host, port) if port else host

            # dialect+driver://username:password@servername:port/database
            connection = '{}+{}://{}:{}@{}/{}'.format(
                database_information['dialect'],
                database_information['driver'],
                database_information['username'],
                database_information['password'],
                host,
                database_information['database'])
            self.engine = create_engine(connection)
            # self.engine.raw_connection()
            self.output_dir = self.output_dir.replace('/', '_')

        elif self.output_type == 'csv':
            self.counts_path = str(Path(self.output_dir, self.output_file))
        else:
            print('unknown output type !!!{]!!!')

    def update_outputs(self, df=None):
        """Update outputs: writing in csv file or in database

        """
        if self.output_type == 'csv':
            header = not exists(self.counts_path)
            with open(self.counts_path, 'a') as f:
                df.to_csv(f, header=header, index=False)
        elif self.output_type == 'database':
            df.to_sql(self.output_dir, self.engine, if_exists='append',
                      index=False, chunksize=10000)


#           _         _                  _    _____ _                 _
#     /\   | |       | |                | |  / ____(_)               | |
#    /  \  | |__  ___| |_ _ __ __ _  ___| |_| (___  _ _ __ ___  _   _| | __ _
#   / /\ \ | '_ \/ __| __| '__/ _` |/ __| __|\___ \| | '_ ` _ \| | | | |/ _` |
#  / ____ \| |_) \__ \ |_| | | (_| | (__| |_ ____) | | | | | | | |_| | | (_| |
# /_/    \_\_.__/|___/\__|_|  \__,_|\___|\__|_____/|_|_| |_| |_|\__,_|_|\__,_|

class AbstractSimulation(object):
    """Abstract class from which any simulation class inherits.

    """
    def __init__(self, start_id: int = 0, model=None, model_path: str = '',
                 stock_agent: bool = True, output_dir: str = 'outputs/',
                 target_agent_class=None, save_results: bool = True,
                 input_dir: str=None,
                 load_from_file: str = None, save_to_file: str = None, **_):
        """Initialize the simulation.

        Args:
            start_id: ID of the (first) simulation
            model: instance of the model to run
            model_path: path to the filename holding the description of the
              model, used if *model* is None
            stock_agent: TODO
            output_dir: name of the directory for simulation outputs
            target_agent_class: agent class representing the top level in the
              simulation
            save_results: True if simulation outputs have to be saved,
              False otherwise. TODO: should be removed, and replaced
              by a set of OutputManagers dedicated to the specific
              expected outputs.
            load_from_file: a filename from which the initial state of
              the simulation (agents corresponding to levels with
              their state) is read (instead of running the
              `initialize_level` method)
            save_to_file: a filename in which the final state of the
              simulation (agents corresponding to levels with their
              state) is written (after running the `finalize_level`
              method)

        """
        # ID of simulation
        self.start_id = start_id
        self.model = EmulsionModel(filename=model_path) if model is None else model
        self.model.input_dir = input_dir
        self.stock_agent = stock_agent
        self.target_agent_class = target_agent_class
        self.output_manager = OutputManager(model=self.model, output_dir=output_dir)
        self.save_results = save_results
        self.save_to_file = save_to_file
        self.load_from_file = load_from_file

        # filename = 'counts.csv'
        # self.counts_path = str(Path(output_dir, filename))

    @abc.abstractmethod
    def evolve(self, steps: int = 1):
        """Operations to perform at each time step. Should be defined in
        subclasses.

        Args:
            steps:

        """
        pass

    @abc.abstractmethod
    def run(self):
        """Entry point to simulation execution. Should be defined in
        subclasses.

        """
        pass

    def update_csv_counts(self, df, dparams: dict = {}):
        """Update the CSV recording of populations in each state. """
        if self.save_results:
            for name, value in dparams.items():
                df.insert(0, name, value)
            # with open(self.counts_path, 'a') as f:
            #     df.to_csv(f, header=header, index=False)
            self.output_manager.update_outputs(df=df)

    # def counts_to_csv(self):
    #     """Record the counts into a CSV file."""
    #     self.counts.to_csv(self.counts_path, index=False)

#   _____ _                 _       _   _
#  / ____(_)               | |     | | (_)
# | (___  _ _ __ ___  _   _| | __ _| |_ _  ___  _ __
#  \___ \| | '_ ` _ \| | | | |/ _` | __| |/ _ \| '_ \
#  ____) | | | | | | | |_| | | (_| | |_| | (_) | | | |
# |_____/|_|_| |_| |_|\__,_|_|\__,_|\__|_|\___/|_| |_|

class Simulation(AbstractSimulation):
    """Simulation class is aimed at running one repetition of a given
    model (for several repetitions, use `MultiSimulation`).

    """
    def __init__(self, steps: int = 100, simu_id: int = 0,
                 silent: bool = False, quiet: bool = False, **others):
        """Create an instance of simulation.

        Args:
            steps: number of time steps to run
            simu_id: ID of the simulation
            silent: if False, show a progress bar during simulation execution
            quiet: if True, show no progressbar at all

        See Also:
            `emulsion.tools.simulation.AbstractSimulation`_
        """
        super().__init__(**others)
        self.simu_id = simu_id
        self.steps = steps
        self.silent = silent
        self.quiet = quiet

        del(others['target_agent_class'])
        others['simu_id'] = self.simu_id
        others['simulation'] = self
        self.agent = self.init_agent(**others) ##self.target_agent_class(**others)

        self.outputs_period = self.model.outputs[self.agent.level]['period']\
                                if self.agent.level in self.model.outputs else 1

    def log_path(self):
        """Return the log path used by current simulation"""
        return self.output_manager.log_path

    def init_agent(self, **others):
        """Create an agent from the target class."""
        toplevel_agent = self.target_agent_class(**others)
        # if self.load_from_file:
        #     toplevel_agent.load_state_from_file(self.simu_id, self.load_from_file)
        # else:
        #     toplevel_agent.initialize_level()
        # # print(toplevel_agent.statevars)
        return toplevel_agent

    def evolve(self, steps: int = 1):
        """Make the target agent evolve.

        Args:
            steps: the number of time steps to run
        """
        for _ in range(steps):
            self.agent.evolve()

    def run(self, dparams: dict = {}):
        """Make the simulation progress."""
        if self.silent or self.quiet:
            progress = range(self.steps)
        else:
            progress = trange(self.steps, desc=f'[Run {self.simu_id}]',
                              bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}')
        self.update_csv_counts(self.counts, dparams=dparams)
        for step in progress:
            self.agent.evolve()
            if step % self.outputs_period == 0:
                self.update_csv_counts(self.counts,
                                       # header=not exists(self.counts_path),
                                       dparams=dparams)

        ## make main level agent run a "finalization" procedure if any
        self.agent.finalize_level(simulation=self)
        if self.save_to_file:
            self.agent.save_state_to_file(self.simu_id, self.save_to_file)

    @property
    def counts(self):
        """Return a pandas DataFrame contains counts of each process if existing.
        TODO: column steps need to be with one of process
        and NO column steps for inter herd

        """
        res = self.agent.counts
        # insert explcitily modified parameters if asked for log
        to_log = self.model.params_to_log
        for name in sorted(to_log.keys()):
            res.insert(0, name, to_log[name])
        # insert ID for the simulation
        res.insert(0, 'simu_id', self.simu_id)
        return res
        # res = None
        # for comp in self.agent:
        #     try:
        #         counts = pd.DataFrame(comp.counts)
        #         res = res.join(counts, lsuffix='res', rsuffix='counts')\
        #                                         if not res is None else counts
        #     except AttributeError:
        #         pass
        #     except Exception as e:
        #         raise e
        # if not res is None:
        #     res.insert(0, 'steps', res.index)
        #     res.insert(0, 'simu_id', self.simu_id)
        # return res

#  __  __       _ _   _  _____ _                 _       _   _
# |  \/  |     | | | (_)/ ____(_)               | |     | | (_)
# | \  / |_   _| | |_ _| (___  _ _ __ ___  _   _| | __ _| |_ _  ___  _ __
# | |\/| | | | | | __| |\___ \| | '_ ` _ \| | | | |/ _` | __| |/ _ \| '_ \
# | |  | | |_| | | |_| |____) | | | | | | | |_| | | (_| | |_| | (_) | | | |
# |_|  |_|\__,_|_|\__|_|_____/|_|_| |_| |_|\__,_|_|\__,_|\__|_|\___/|_| |_|

class MultiSimulation(AbstractSimulation):
    """MultiSimulation can handle multiple repetitions of a given model.
    For sensibility study (same model with different values of variables),
    please check out SensitivitySimulation.

    """
    def __init__(self, multi_id: int = 0, nb_simu: int = 100,
                 set_seed: bool = False, silent: bool = False,
                 quiet: bool = False, dparams: dict = {}, **others):
        """Initialize a simulation with multiple repetitions of the same
        model.
        """
        super().__init__(**others)
        self.multi_id = multi_id
        self.nb_simu = nb_simu
        self.set_seed = set_seed
        self.others = others
        self.silent = silent
        self.quiet = quiet
        self.dparams = dparams

        self.d_simu = dict()

    def __iter__(self):
        return self.d_simu.values().__iter__()

    def evolve(self, steps=1):
        for simu in self.d_simu.values():
            simu.evolve(steps=steps)

    def run(self, update=True):
        """Run all repetitions one by one.

        """
        if self.silent and not self.quiet:
            progress = tqdm(range(self.start_id, self.start_id+self.nb_simu),
                            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}')
        else:
            progress = range(self.start_id, self.start_id+self.nb_simu)
        for simu_id in progress:
            if self.set_seed:
                np.random.seed(simu_id)
            if 'model' not in self.others:
                self.others['model'] = self.model
            simu = Simulation(simu_id=simu_id, silent=self.silent,
                              quiet=self.quiet, dparams=self.dparams,
                              **self.others)

            simu.run()

            if self.stock_agent:
                self.d_simu[simu_id] = simu

            # if update:
            #     header = not exists(self.counts_path)
            #     self.update_csv_counts(simu.counts, header=header, dparams=dparams)

    @property
    def counts(self):
        l_counts = [simu.counts for simu in self]
        return pd.concat(l_counts).reset_index(drop=True)

    def write_dot(self):
        self.model.write_dot(self.others['output_dir'])


#   _____                _ _   _       _ _          _____ _                 _
#  / ____|              (_) | (_)     (_) |        / ____(_)               | |
# | (___   ___ _ __  ___ _| |_ ___   ___| |_ _   _| (___  _ _ __ ___  _   _| |
#  \___ \ / _ \ '_ \/ __| | __| \ \ / / | __| | | |\___ \| | '_ ` _ \| | | | |
#  ____) |  __/ | | \__ \ | |_| |\ V /| | |_| |_| |____) | | | | | | | |_| | |
# |_____/ \___|_| |_|___/_|\__|_| \_/ |_|\__|\__, |_____/|_|_| |_| |_|\__,_|_|
#                                             __/ |
#                                            |___/
#        _   _
#       | | (_)
#   __ _| |_ _  ___  _ __
#  / _` | __| |/ _ \| '_ \
# | (_| | |_| | (_) | | | |
#  \__,_|\__|_|\___/|_| |_|

class SensitivitySimulation(AbstractSimulation):
    """SensitivitySimulation can handle sensibility study with a given
    pandas DataFrame of parameters or a path linked with file which contains
    scenarios of parameters. Then it will be transformed to a dictionary of
    scenario in the ```d_scenario``` attribute.

    For instance, d_scenario could be the form (QFever model example) :
        {0: {'m': 0.7, 'q': 0.02 ...},
         1: {'m': 0.5, 'q': 0.02 ...},
         2: ...,
         ... }
    """
    def __init__(self, scenario_path=None, df=None, nb_multi=None, **others):
        super().__init__(**others)
        self.others = others
        self.others['start_id'] = 0

        # Retrieving DataFrame and creation of dictionnary of simulation/scenario
        df = pd.read_csv(scenario_path) if df is None else df
        self.nb_multi = len(df) if nb_multi is None else nb_multi
        self.d_scenario = df.to_dict(orient='index')
        self.d_multi = dict()

    def run(self):
        """Make the simulation advance."""
        bar_sens = tqdm(range(self.start_id, self.start_id+self.nb_multi),
                        desc='[Sensitivity Simulation]')

        # for multi_id, scenario in bar_sens(self.d_scenario.items()):
        for multi_id in bar_sens:
            scenario = self.d_scenario[multi_id]
            # Copy a model
            model = self.model.copy()
            # Modify model
            for name, value in scenario.items():
                try:
                    model.set_value(name, value)
                except Exception:
                    self.others[name] = value

            # Instantiate MultiSimulation and execution
            multi = MultiSimulation(multi_id=multi_id, model=model, **self.others)
            multi.run(dparams=scenario)

            if self.stock_agent:
                self.d_multi[multi_id] = multi

            # if update:
            #     header = not exists(self.counts_path)

            #     counts = multi.counts
            #     for name, value in self.d_scenario[multi_id].items():
            #         counts.insert(0, name, value)

            #     self.update_csv_counts(counts, header=header)

    @property
    def counts(self):
        l_counts = []
        for multi_id, multi in self.d_multi.items():
            counts = multi.counts
            for name, value in self.d_scenario[multi_id].items():
                counts.insert(0, name, value)

            counts.insert(0, 'scenario_id', multi_id)
            l_counts.append(counts)
        print(counts)
        return pd.concat(l_counts).reset_index(drop=True)

    def write_dot(self):
        self.model.write_dot(self.others['output_dir'])
